from __future__ import annotations

from copy import copy
from typing import TYPE_CHECKING, cast

from pdm.models.repositories import BaseRepository
from pdm.models.requirements import strip_extras
from pdm.resolver.providers import (
    BaseProvider,
    EagerUpdateProvider,
    ReuseInstalledProvider,
    ReusePinProvider,
    register_provider,
)
from pdm.utils import is_url
from unearth.utils import LazySequence

from pdm_conda.conda import CondaResolutionError
from pdm_conda.environments import CondaEnvironment
from pdm_conda.models.candidates import CondaCandidate
from pdm_conda.models.repositories import CondaRepository
from pdm_conda.models.requirements import CondaRequirement, as_conda_requirement, parse_requirement

if TYPE_CHECKING:
    from collections.abc import Callable, Iterable, Iterator, Mapping, Sequence

    from pdm.models.candidates import Candidate
    from pdm.models.requirements import Requirement
    from pdm.resolver.providers import Comparable
    from resolvelib.resolvers import RequirementInformation


@register_provider("all")
class CondaBaseProvider(BaseProvider):
    def __init__(
        self,
        repository: BaseRepository,
        allow_prereleases: bool | None = None,
        overrides: dict[str, str] | None = None,
        direct_minimal_versions: bool = False,
        locked_candidates: dict[str, Candidate] | None = None,
    ) -> None:
        super().__init__(repository, allow_prereleases, overrides, direct_minimal_versions, locked_candidates)
        self._overrides_requirements: dict | None = None
        environment = repository.environment
        self.is_conda_initialized = (
            isinstance(environment, CondaEnvironment) and environment.project.conda_config.is_initialized
        )
        if self.is_conda_initialized:
            self.excludes = {
                parse_requirement(name).identify()
                for name in environment.project.pyproject.resolution.get("excludes", [])
            }
        python_version = str(environment.interpreter.version)
        self.python_candidate = CondaCandidate(
            parse_requirement(f"conda:python=={python_version}"),
            "python",
            python_version,
        )

    @property
    def overrides_requirements(self) -> dict[str, Requirement]:
        """Identifier and requirement mapping for overrides.

        :return: mapping
        """
        if self._overrides_requirements is None:
            self._overrides_requirements = {}
            if self.overrides:
                for identifier, requested in self.overrides.items():
                    if is_url(requested):
                        requirement = parse_requirement(f"{identifier} @ {requested}")
                    else:
                        # first parse as conda to ensure no version error
                        requirement = cast(CondaRequirement, parse_requirement(f"conda:{identifier} {requested}"))
                        if not isinstance(self.repository, CondaRepository) or not self.repository.is_conda_managed(
                            requirement,
                        ):
                            requirement.name = identifier
                            requirement = parse_requirement(requirement.as_line())
                    self._overrides_requirements[self.identify(requirement)] = requirement

        return self._overrides_requirements

    def get_preference(
        self,
        identifier: str,
        resolutions: dict[str, Candidate],
        candidates: dict[str, Iterator[Candidate]],
        information: dict[str, Iterator[RequirementInformation]],
        backtrack_causes: Sequence[RequirementInformation],
    ) -> tuple[Comparable, ...]:
        preference = super().get_preference(identifier, resolutions, candidates, information, backtrack_causes)
        if self.is_conda_initialized:
            return (
                preference[:3],
                (
                    isinstance(self.repository, CondaRepository)
                    and self.repository.is_conda_managed(next(information[identifier]).requirement)
                ),
                *preference[3:],
            )
        return preference

    def find_matches(
        self,
        identifier: str,
        requirements: Mapping[str, Iterator[Requirement]],
        incompatibilities: Mapping[str, Iterator[Candidate]],
    ) -> Callable[[], Iterator[Candidate]]:
        super_find_matches = super().find_matches(identifier, requirements, incompatibilities)
        if not self.is_conda_initialized:
            return super_find_matches

        def matches_gen() -> Iterator[Candidate]:
            incompat = list(incompatibilities[identifier])
            bare_name, extras = strip_extras(identifier)
            if identifier == "python" or any(name in self.overrides for name in (identifier, bare_name)):
                return super_find_matches()
            reqs = sorted(requirements[identifier], key=self.requirement_preference)
            if not reqs:
                return iter(())
            original_req = reqs[0]
            if extras and bare_name in requirements:
                # We should consider the requirements for both foo and foo[extra]
                reqs.extend(requirements[bare_name])
                reqs.sort(key=self.requirement_preference)
            # iterates over requirements
            candidates = []
            for req in reqs:
                candidates = self._find_candidates(req)
                candidates = LazySequence(
                    # In some cases we will use candidates from the bare requirement,
                    # this will miss the extra dependencies if any. So we associate the original
                    # requirement back with the candidate since it is used by `get_dependencies()`.
                    (
                        (
                            can.copy_with(original_req, merge_requirements=True)
                            if isinstance(can, CondaCandidate)
                            else can.copy_with(original_req)
                        )
                        if extras
                        else can
                    )
                    for can in candidates
                    if can not in incompat and all(self.is_satisfied_by(r, can) for r in reqs)
                )
                if candidates:
                    break
            return iter(candidates)

        return matches_gen

    def get_requirement_from_overrides(self, requirement: Requirement) -> Requirement:
        _req = copy(self.overrides_requirements.get(self.identify(requirement), requirement))
        if not requirement.groups:
            _req.groups = requirement.groups
        if isinstance(requirement, CondaRequirement):
            _req = as_conda_requirement(_req)
        return _req

    def get_override_candidates(self, identifier: str) -> Iterable[Candidate]:
        if self.is_conda_initialized:
            return self._find_candidates(self.overrides_requirements[identifier])
        return super().get_override_candidates(identifier)

    def compatible_with_resolution(
        self,
        requirements: list[Requirement],
        resolution: dict,
        excluded_identifiers: set[str],
    ) -> bool:
        """True if all requirements are compatible with the resolution.

        :param requirements: list of requirements
        :param resolution: resolution to check
        :param excluded_identifiers: identifiers to exclude
        :return: True if all requirements are compatible with the resolution
        """
        return self.repository.compatible_with_resolution(
            [self.get_requirement_from_overrides(req) for req in requirements],
            resolution,
            excluded_identifiers,
        )

    def update_conda_resolution(
        self,
        new_requirements: list[Requirement] | None = None,
        requirements: list[Requirement] | None = None,
        resolution: dict | None = None,
        excluded_identifiers: set[str] | None = None,
    ) -> set[str]:
        """Updates the existing conda resolution if new requirements.

        :param new_requirements: new requirements to add
        :param requirements: list of requirements
        :param resolution: resolution to override existing
        :param excluded_identifiers: identifiers to exclude
        :return: list of excluded identifiers
        """
        return self.repository.update_conda_resolution(
            [self.get_requirement_from_overrides(req) for req in new_requirements] if new_requirements else None,
            [self.get_requirement_from_overrides(req) for req in requirements] if requirements else None,
            resolution,
            excluded_identifiers,
        )


@register_provider("reuse")
class CondaReusePinProvider(ReusePinProvider, CondaBaseProvider):
    def __init__(self, *args, **kwargs) -> None:
        super().__init__(*args, **kwargs)
        for name in self.tracked_names:
            self.locked_candidates.pop(name, None)

    def find_matches(
        self,
        identifier: str,
        requirements: Mapping[str, Iterator[Requirement]],
        incompatibilities: Mapping[str, Iterator[Candidate]],
    ) -> Callable[[], Iterator[Candidate]]:
        super_find = super(CondaBaseProvider, self).find_matches(identifier, requirements, incompatibilities)

        def matches_gen() -> Iterator[Candidate]:
            requested_req = next(filter(lambda r: r.is_named, requirements[identifier]), None)
            pin = self.get_reuse_candidate(identifier, requested_req)
            if pin is not None:
                incompat = list(incompatibilities[identifier])
                pin._preferred = True  # type: ignore[attr-defined]
                if pin not in incompat and all(self.is_satisfied_by(r, pin) for r in requirements[identifier]):
                    yield pin
            yield from super_find()

        return matches_gen

    def _merge_requirements(
        self,
        requirements: list[Requirement] | None,
        excluded=None,
        include_all: bool = False,
    ) -> list[Requirement]:
        _requirements = []
        excluded = set(excluded or set())
        requirements = requirements or []
        for req in requirements:
            ident = self.identify(req)
            if (
                self.repository.is_conda_managed(req, excluded)
                and ident in self.locked_candidates
                and as_conda_requirement(req).is_compatible(can := self.locked_candidates[ident])
            ):
                _requirements.append(can.req)
            else:
                _requirements.append(req)
            excluded.add(ident)
        if include_all:
            for can in self.locked_candidates.values():
                if self.identify(can.req) not in excluded:
                    _requirements.append(can.req)

        return _requirements

    def update_conda_resolution(
        self,
        new_requirements: list[Requirement] | None = None,
        requirements: list[Requirement] | None = None,
        resolution: dict | None = None,
        excluded_identifiers: set[str] | None = None,
    ) -> set[str]:
        """Updates the existing conda resolution if new requirements, keeping the pinned versions if possible.

        :param new_requirements: new requirements to add
        :param requirements: list of requirements
        :param resolution: resolution to override existing
        :param excluded_identifiers: identifiers to exclude
        :return: list of excluded identifiers
        """
        # try to reuse the pinned versions
        new_requirements = new_requirements or []
        excluded_identifiers = excluded_identifiers or set()
        try:
            return super().update_conda_resolution(
                self._merge_requirements(new_requirements, excluded_identifiers),
                self._merge_requirements(
                    requirements,
                    excluded_identifiers | {req.key for req in new_requirements},
                    include_all=True,
                ),
                resolution,
                excluded_identifiers,
            )
        except CondaResolutionError:
            return super().update_conda_resolution(
                new_requirements,
                requirements,
                resolution,
                excluded_identifiers,
            )


@register_provider("eager")
class CondaEagerUpdateProvider(EagerUpdateProvider, CondaBaseProvider):
    pass


@register_provider("reuse-installed")
class CondaReuseInstalledProvider(ReuseInstalledProvider, CondaBaseProvider):
    pass



from __future__ import annotations

import contextlib
from dataclasses import dataclass, field
from itertools import chain

from resolvelib.resolvers import RequirementInformation, Resolution, Resolver, _build_result

from pdm_conda.conda import CondaResolutionError
from pdm_conda.environments import CondaEnvironment
from pdm_conda.models.candidates import CondaCandidate
from pdm_conda.models.requirements import CondaRequirement, as_conda_requirement

CONSTRAINS_KEY = "___constrains___"
CONDA_RESOLUTION_KEY = "___conda_resolution___"
CONDA_EXCLUDED_IDENTIFIERS_KEY = "___conda_excluded_identifiers___"


@dataclass
class State:
    mapping: dict
    criteria: dict
    backtrack_causes: list
    constrains: dict = field(default_factory=dict)
    conda_resolution: dict = field(default_factory=dict)
    conda_excluded_identifiers: set[str] = field(default_factory=set)


class CondaResolution(Resolution):
    def __init__(
        self,
        provider,
        reporter,
        base_constrains: dict | None = None,
        conda_resolution: dict | None = None,
        conda_excluded_identifiers: set[str] | None = None,
        is_conda_initialized: bool = True,
    ) -> None:
        super().__init__(provider, reporter)
        self._is_conda_initialized = is_conda_initialized
        self._base_constrains = base_constrains or {}
        if not conda_resolution and is_conda_initialized:
            conda_resolution = {
                can.req.conda_name: [can]
                for can in provider.locked_candidates.values()
                if isinstance(can, CondaCandidate)
            }
            conda_resolution["python"] = [provider.python_candidate]
        self._conda_resolution = conda_resolution or {}
        self._conda_excluded_identifiers = conda_excluded_identifiers or set()

    @property
    def state(self):
        if self._states and not isinstance(state := self._states[-1], State):
            self._states[-1] = State(
                state.mapping,
                state.criteria,
                state.backtrack_causes,
                self._base_constrains,
                self._conda_resolution,
                self._conda_excluded_identifiers,
            )
        return super().state

    def _push_new_state(self):
        base = self.state
        base.constrains = base.criteria.pop(CONSTRAINS_KEY, base.constrains)
        base.conda_resolution = base.criteria.pop(CONDA_RESOLUTION_KEY, base.conda_resolution)
        base.conda_excluded_identifiers = base.criteria.pop(
            CONDA_EXCLUDED_IDENTIFIERS_KEY,
            base.conda_excluded_identifiers,
        )
        state = State(
            mapping=base.mapping.copy(),
            criteria=base.criteria.copy(),
            backtrack_causes=base.backtrack_causes[:],
            constrains=base.constrains.copy(),
            conda_resolution=base.conda_resolution.copy(),
            conda_excluded_identifiers=base.conda_excluded_identifiers.copy(),
        )
        # update repository conda resolution to latest
        if self._is_conda_initialized:
            self._p.update_conda_resolution(
                resolution=state.conda_resolution,
                excluded_identifiers=state.conda_excluded_identifiers,
            )
        self._states.append(state)

    def _remove_information_from_criteria(self, criteria, parents):
        state = self.state
        state.constrains = state.criteria.pop(CONSTRAINS_KEY, state.constrains)
        state.conda_resolution = state.criteria.pop(CONDA_RESOLUTION_KEY, state.conda_resolution)
        state.conda_excluded_identifiers = state.criteria.pop(
            CONDA_EXCLUDED_IDENTIFIERS_KEY,
            state.conda_excluded_identifiers,
        )
        super()._remove_information_from_criteria(criteria, parents)

    def _ensure_criteria(self, criteria):
        # save temporarily constrains and conda_resolution in criteria
        for k, v in [
            (CONDA_RESOLUTION_KEY, self.state.conda_resolution),
            (CONSTRAINS_KEY, self.state.constrains),
            (CONDA_EXCLUDED_IDENTIFIERS_KEY, self.state.conda_excluded_identifiers),
        ]:
            if k not in criteria:
                criteria[k] = v.copy()

    def update_constrains(self, candidate: CondaCandidate, criteria=None, merge_old: bool | set[str] = True):
        if isinstance(merge_old, bool):
            merge_old = set(candidate.constrains.keys()) if merge_old else set()
        if criteria is not None:
            self._ensure_criteria(criteria)
            constrains = criteria[CONSTRAINS_KEY]
        else:
            constrains = self._base_constrains

        for identifier, constrain in candidate.constrains.items():
            if identifier != "python":
                # keep most restrictive constrain
                if identifier in merge_old and (existing_constrain := constrains.get(identifier, None)) is not None:
                    constrain = existing_constrain.merge(constrain)
                constrains[identifier] = constrain
                if criteria is not None and identifier in criteria:
                    self._add_to_criteria(criteria, constrain, parent=candidate)

    def _update_conda_resolution(self, criteria, new_requirements):
        # update conda resolution with new requirements
        self._ensure_criteria(criteria)
        resolution = criteria[CONDA_RESOLUTION_KEY]
        excluded_identifiers = criteria[CONDA_EXCLUDED_IDENTIFIERS_KEY]
        if not self._p.compatible_with_resolution(new_requirements, resolution, excluded_identifiers):
            requirements = list(
                chain.from_iterable(
                    (
                        [information.requirement for information in criterion.information]
                        for i, criterion in criteria.items()
                        if i not in [CONDA_RESOLUTION_KEY, CONSTRAINS_KEY, CONDA_EXCLUDED_IDENTIFIERS_KEY]
                        and criterion.information
                    ),
                ),
            )
            with contextlib.suppress(CondaResolutionError):
                criteria[CONDA_EXCLUDED_IDENTIFIERS_KEY] = self._p.update_conda_resolution(
                    new_requirements,
                    requirements,
                    resolution=resolution,
                    excluded_identifiers=excluded_identifiers,
                )

    def _add_to_criteria(self, criteria, requirement, parent):
        if self._is_conda_initialized:
            # merge with constrain if exists
            self._ensure_criteria(criteria)
            constrains = criteria[CONSTRAINS_KEY]
            if (constrain := constrains.get(requirement.conda_name, None)) is not None:
                requirement = constrain.merge(requirement)

            self._update_conda_resolution(criteria, [requirement])
            if criterion := criteria.get(self._p.identify(requirement)):
                # if excluded then delete conda related information else if other conda requirement transform to conda
                if not self._p.repository.is_conda_managed(requirement, criteria[CONDA_EXCLUDED_IDENTIFIERS_KEY]):
                    criterion.information = [
                        (
                            RequirementInformation(i.requirement.as_named_requirement(), i.parent)
                            if isinstance(i.requirement, CondaRequirement)
                            else i
                        )
                        for i in criterion.information
                    ]
                # if not excluded and conda requirement then transform related information to conda
                elif isinstance(requirement, CondaRequirement):
                    criterion.information = [
                        RequirementInformation(as_conda_requirement(i.requirement), i.parent)
                        for i in criterion.information
                    ]

        super()._add_to_criteria(criteria, requirement, parent)

    def _get_updated_criteria(self, candidate):
        criteria = self.state.criteria.copy()
        self._ensure_criteria(criteria)
        dependencies = self._p.get_dependencies(candidate=candidate)
        # update conda resolution with dependencies if parent excluded
        if self._is_conda_initialized and not self._p.repository.is_conda_managed(
            candidate.req,
            criteria[CONDA_EXCLUDED_IDENTIFIERS_KEY],
        ):
            self._update_conda_resolution(criteria, dependencies)

        for requirement in dependencies:
            self._add_to_criteria(criteria, requirement, parent=candidate)

        # merge with previous constrain if exists
        if self._is_conda_initialized and isinstance(candidate, CondaCandidate):
            self.update_constrains(candidate, criteria)
        return criteria

    def initialize_conda_resolution(self, requirements, excluded_identifiers: set[str] | None):
        # update conda resolution
        self._conda_excluded_identifiers = self._p.update_conda_resolution(
            requirements
            if not self._p.compatible_with_resolution(requirements, self._conda_resolution, excluded_identifiers)
            else None,
            resolution=self._conda_resolution,
            excluded_identifiers=excluded_identifiers,
        )

        # update constrains
        for candidates in self._conda_resolution.values():
            for can in candidates:
                self.update_constrains(candidate=can)


class CondaResolver(Resolver):
    def resolve(self, requirements, max_rounds=100):
        project = self.provider.repository.environment.project
        is_conda_initialized = (
            isinstance(self.provider.repository.environment, CondaEnvironment) and project.conda_config.is_initialized
        )
        resolution = CondaResolution(self.provider, self.reporter, is_conda_initialized=is_conda_initialized)
        if is_conda_initialized:
            conda_config = project.conda_config
            resolution.initialize_conda_resolution(requirements, conda_config.excluded_identifiers)
            if conda_config.custom_behavior:
                project.is_distribution = True
        else:
            for r in requirements:
                if isinstance(r, CondaRequirement):
                    raise CondaResolutionError(f"Conda requirement {r} detected but pdm-plugin is not initialized.")

        try:
            state = resolution.resolve(requirements, max_rounds=max_rounds)
            result = _build_result(state)
        finally:
            if is_conda_initialized:
                project.is_distribution = None

        # here we remove a self dependency we added because of the custom behavior
        if is_conda_initialized:
            if conda_config.custom_behavior and not project.is_distribution:
                for key, candidate in list(result.mapping.items()):
                    if candidate.name == conda_config.project_name:
                        del result.mapping[key]
            if conda_config.auto_excludes:
                conda_config.excludes = state.conda_excluded_identifiers
        return result



from __future__ import annotations

import sys
from typing import TYPE_CHECKING

from findpython.providers.base import BaseProvider
from pdm.cli.commands.venv import list, utils
from pdm.models.venv import VirtualEnv

from pdm_conda.conda import conda_env_list
from pdm_conda.models.config import PluginConfig
from pdm_conda.utils import get_python_dir

if TYPE_CHECKING:
    from pdm_conda.project import CondaProject
    from typing_extensions import Self
    from typing import Iterable
    from findpython.python import PythonVersion

get_venv_prefix = utils.get_venv_prefix


def find_pythons(project) -> Iterable[PythonVersion]:
    python_suffix = "bin/python" if sys.platform != "win32" else "python.exe"
    for env in conda_env_list(project):
        if env != project.base_env and env.parent != project.base_env.parent:
            python_bin = env / python_suffix
            if python_bin.exists():
                yield CondaProvider.version_maker(
                    python_bin,
                    _interpreter=python_bin,
                    keep_symlink=False,
                )


class CondaProvider(BaseProvider):
    """A provider that finds python installed with Conda."""

    def __init__(self, project: CondaProject) -> None:
        super().__init__()
        self.project = project

    @classmethod
    def create(cls) -> Self | None:
        return None

    def find_pythons(self) -> Iterable[PythonVersion]:
        yield from find_pythons(self.project)


def wrap_iter_venvs(func):
    @PluginConfig.check_active
    def wrapper(project):
        yield from func(project)
        if project.conda_config.is_initialized:
            for python in find_pythons(project):
                venv = VirtualEnv.get(get_python_dir(python.executable))
                if venv.is_conda:
                    yield venv.root.name, venv

    return wrapper


for module in [utils, list]:
    module.iter_venvs = wrap_iter_venvs(module.iter_venvs)



from __future__ import annotations

import os
from typing import TYPE_CHECKING, cast

from pdm.cli.commands.use import Command as BaseCommand
from pdm.utils import is_conda_base

from pdm_conda.cli.utils import ensure_logger
from pdm_conda.models.config import PluginConfig
from pdm_conda.project import CondaProject

if TYPE_CHECKING:
    import argparse

    from pdm_conda.project import Project


class Command(BaseCommand):
    description = BaseCommand.__doc__
    name = "use"

    @PluginConfig.check_active
    def handle(self, project: Project, options: argparse.Namespace) -> None:
        project = cast(CondaProject, project)
        conda_base = is_conda_base()
        conda_default_env = ""
        if project.conda_config.is_initialized and conda_base:
            conda_default_env = os.environ.pop("CONDA_DEFAULT_ENV", "")
        try:
            with ensure_logger(project, "use"):
                super().handle(project, options)
        finally:
            if project.conda_config.is_initialized and conda_base:
                os.environ["CONDA_DEFAULT_ENV"] = conda_default_env



from __future__ import annotations

from typing import cast, TYPE_CHECKING

from pdm.cli.commands.venv.create import CreateCommand as BaseCommand
from pdm.cli.options import ArgumentGroup, split_lists

from pdm_conda.models.config import CondaRunner

if TYPE_CHECKING:
    import argparse
    from pdm.project import Project


class CreateCommand(BaseCommand):
    description = BaseCommand.__doc__

    def add_arguments(self, parser: argparse.ArgumentParser) -> None:
        super().add_arguments(parser)
        conda_group = ArgumentGroup("Conda Options")
        conda_group.add_argument(
            "-c",
            "--channel",
            dest="conda_channels",
            metavar="CHANNEL",
            action=split_lists(","),
            help="Specify Conda channels separated by comma, can be supplied multiple times",
            default=[],
        )
        conda_group.add_argument(
            "-cn",
            "--conda-name",
            help="Specify the name of the Conda environment, overrides --name and appended hash",
        )
        conda_group.add_to_parser(parser)

    def handle(self, project: Project, options: argparse.Namespace) -> None:
        from pdm_conda.project import CondaProject

        default_backend = project.config["venv.backend"]
        project = cast(CondaProject, project)
        conda_config = project.conda_config
        backend = options.backend or default_backend
        if conda_config.is_initialized and backend == default_backend and conda_config.custom_behavior:
            backend = options.backend = conda_config.runner

        overridden_configs = {}
        if conda_project := (backend in list(CondaRunner)):
            overridden_configs["is_initialized"] = True
            overridden_configs["runner"] = backend
            conda_config.runner = backend
            conda_config.is_initialized = True
            if options.conda_name:
                options.name = f"conda:{options.conda_name}"
            channels = options.conda_channels
            for channel in channels:
                if channel not in conda_config.channels:
                    conda_config.channels.append(channel)
            overridden_configs["channels"] = conda_config.channels

        conda_config.check_active(super().handle)(project, options)

        # if inside a project ensure saving conda runner
        if conda_project and project.pyproject.exists():
            with conda_config.write_project_config():
                for key, value in overridden_configs.items():
                    setattr(conda_config, key, value)



from __future__ import annotations

import contextlib
import functools
from typing import TYPE_CHECKING

from pdm.cli import actions, utils
from pdm.formats.base import array_of_inline_tables, make_array
from pdm.models.specifiers import get_specifier

from pdm_conda import logger
from pdm_conda.models.candidates import CondaCandidate
from pdm_conda.models.repositories import CondaRepository
from pdm_conda.models.requirements import CondaRequirement, as_conda_requirement, comparable_version

if TYPE_CHECKING:
    from collections.abc import Mapping

    from pdm.project import Project

    from pdm_conda.models.candidates import Candidate
    from pdm_conda.models.requirements import Requirement


@contextlib.contextmanager
def ensure_logger(project, logger_name: str):
    if len(logger.handlers) == 1:
        with project.core.ui.logging(logger_name):
            yield
    else:
        yield


def remove_quotes(req: str) -> str:
    for quote in ("'", '"'):
        if req.startswith(quote) and req.endswith(quote):
            req = req[1:-1]
    return req


def wrap_fetch_hashes(func):
    @functools.wraps(func)
    def wrapper(repository, mapping: Mapping[str, Candidate]) -> None:
        conda_candidates = {}
        if isinstance(repository, CondaRepository):
            conda_candidates = {name: can for name, can in mapping.items() if isinstance(can, CondaCandidate)}
            repository.update_hashes(conda_candidates)

        return func(repository, {name: can for name, can in mapping.items() if name not in conda_candidates})

    return wrapper


def wrap_save_version_specifiers(func):
    @functools.wraps(func)
    def wrapper(
        requirements: dict[str, dict[str, Requirement]],
        resolved: dict[str, Candidate],
        save_strategy: str,
    ) -> None:
        func(requirements, resolved, save_strategy)
        for reqs in requirements.values():
            for name, r in reqs.items():
                can = resolved[name]
                if save_strategy == "compatible" and r.is_named and (version := comparable_version(can.version)).epoch:
                    if version.is_prerelease or version.is_devrelease:
                        r.specifier = get_specifier(
                            f">={version.epoch}!{version},<{version.epoch}!{version.major + 1}",
                        )
                    else:
                        r.specifier = get_specifier(f"~={version.epoch}!{version.major}.{version.minor}")
                if isinstance(can, CondaCandidate):
                    r = as_conda_requirement(r)
                    r.version_mapping.update(can.req.version_mapping)
                    r.is_python_package = can.req.is_python_package
                    reqs[name] = r

    return wrapper


def wrap_format_lockfile(func):
    @functools.wraps(func)
    def wrapper(
        project: Project,
        mapping: dict[str, Candidate],
        fetched_dependencies: dict[tuple[str, str | None], list[Requirement]],
        *args,
        **kwargs,
    ) -> dict:
        res = func(project, mapping, fetched_dependencies, *args, **kwargs)
        # ensure no duplicated groups in metadata
        if groups := res.get("metadata", {}).get("groups"):
            res["metadata"]["groups"] = list({group: None for group in groups}.keys())

        assert len(res["package"]) == len(mapping)
        # fix conda packages
        for package, (_, can) in zip(res["package"], sorted(mapping.items()), strict=False):
            # only static-url allowed for conda packages
            if isinstance(can, CondaCandidate):
                package["files"] = array_of_inline_tables(
                    [{"url": item["url"], "hash": item["hash"]} for item in can.hashes],
                    multiline=True,
                )

            # fix conda dependencies to include build string
            dependencies = []
            include_dependencies = False
            for dep in fetched_dependencies.get(can.dep_key, []):
                kwargs = {}
                if isinstance(dep, CondaRequirement):
                    kwargs["with_build_string"] = True
                    include_dependencies = True
                dependencies.append(dep.as_line(**kwargs))
            if include_dependencies:
                package["dependencies"] = make_array(sorted(set(dependencies)), True)

        res["package"] = sorted(res["package"], key=lambda x: x["name"])
        return res

    return wrapper


save_version_specifiers = wrap_save_version_specifiers(utils.save_version_specifiers)
format_lockfile = wrap_format_lockfile(utils.format_lockfile)
wrap_fetch_hashes = wrap_fetch_hashes(actions.fetch_hashes)
for m in [utils, actions]:
    m.save_version_specifiers = save_version_specifiers
    m.format_lockfile = format_lockfile
    m.fetch_hashes = wrap_fetch_hashes



from __future__ import annotations

from functools import cached_property
from typing import TYPE_CHECKING

from pdm.compat import Distribution
from pdm.installers import Synchronizer

from pdm_conda.environments import CondaEnvironment
from pdm_conda.installers.manager import CondaInstallManager
from pdm_conda.models.candidates import CondaCandidate
from pdm_conda.models.requirements import strip_extras
from pdm_conda.models.setup import CondaSetupDistribution

if TYPE_CHECKING:
    from collections.abc import Collection

    from pdm_conda.environments import BaseEnvironment
    from pdm_conda.models.candidates import Candidate


class CondaSynchronizer(Synchronizer):
    def __init__(
        self,
        candidates: dict[str, Candidate],
        environment: BaseEnvironment,
        clean: bool = False,
        dry_run: bool = False,
        retry_times: int = 1,
        install_self: bool = False,
        no_editable: bool | Collection[str] = False,
        reinstall: bool = False,
        only_keep: bool = False,
        fail_fast: bool = False,
        use_install_cache: bool | None = None,
    ) -> None:
        super().__init__(
            candidates,
            environment,
            clean,
            dry_run,
            retry_times,
            install_self,
            no_editable,
            reinstall,
            only_keep,
            fail_fast,
            use_install_cache,
        )
        self.parallel = bool(self.parallel)  # type: ignore

    @cached_property
    def candidates(self) -> dict[str, Candidate]:
        candidates = super().candidates
        # if key is requirement with extras, add candidate without extras if it doesn't exist
        for key in list(candidates):
            if (
                isinstance(can := candidates[key], CondaCandidate)
                and candidates.get(name := strip_extras(key)[0], None) is None
            ):
                candidates[name] = can
        return candidates

    def _should_update(self, dist: Distribution, can: Candidate) -> bool:
        # Always update if one of the two is conda and the other is not
        # or as line is different
        return (
            super()._should_update(dist, can)
            or (
                isinstance(can, CondaCandidate)
                and (not isinstance(dist, CondaSetupDistribution) or dist.as_line() != can.req.as_line())
            )
            or (isinstance(dist, CondaSetupDistribution) and not isinstance(can, CondaCandidate))
        )

    def compare_with_working_set(self) -> tuple[list[str], list[str], list[str]]:
        to_add, to_update, to_remove = super().compare_with_working_set()
        if not isinstance(self.environment, CondaEnvironment):
            return to_add, to_update, to_remove
        if to_remove:
            to_remove = [p for p in to_remove if p not in self.environment.env_dependencies]

        if (
            not isinstance(self.manager, CondaInstallManager)
            or not self.environment.project.conda_config.batched_commands
        ):
            return to_add, to_update, to_remove

        to_batch_install = set()
        to_batch_remove = set()

        def extract_conda(pkgs, candidates, cls):
            conda_pkgs = []
            for pkg in pkgs:
                if isinstance(candidates[pkg], cls):
                    conda_pkgs.append(pkg)
            return conda_pkgs

        for pkgs in (to_add, to_update):
            to_batch_install.update(extract_conda(pkgs, self.candidates, CondaCandidate))

        for pkgs in (to_remove, to_update):
            to_batch_remove.update(extract_conda(pkgs, self.working_set, CondaSetupDistribution))

        self.manager.prepare_batch_operations(to_batch_install, to_batch_remove)

        return to_add, to_update, to_remove



from __future__ import annotations

import functools
from typing import TYPE_CHECKING

from pdm.resolver import core, graph
from pdm.resolver.graph import _identify_parent

if TYPE_CHECKING:
    from pdm.models.candidates import Candidate
    from pdm.models.requirements import Requirement
    from resolvelib.resolvers import Result


def wrapper_populate_groups(func):
    @functools.wraps(func)
    def wrapper(result: Result[Requirement, Candidate, str]) -> None:
        """Correct groups for cyclic dependencies."""

        func(result)
        for k, can in reversed(result.mapping.items()):
            groups = set(can.req.groups)
            for _, parent in result.criteria[k].information:
                if (
                    parent is not None
                    and (parent_can := result.mapping.get(_identify_parent(parent), None)) is not None
                ):
                    groups.update(parent_can.req.groups)
            can.req.groups = sorted(groups)

    return wrapper


populate_groups = wrapper_populate_groups(graph.populate_groups)

for module in (graph, core):
    module.populate_groups = populate_groups



from __future__ import annotations

import uuid
from typing import TYPE_CHECKING, cast

from pdm.exceptions import CandidateNotFound
from pdm.models.repositories import BaseRepository, LockedRepository, PyPIRepository
from pdm.models.specifiers import PySpecSet

from pdm_conda import logger
from pdm_conda.conda import CondaResolutionError, CondaSearchError, conda_create, conda_search, sort_candidates
from pdm_conda.environments import CondaEnvironment
from pdm_conda.models.candidates import CondaCandidate
from pdm_conda.models.requirements import CondaRequirement, as_conda_requirement

if TYPE_CHECKING:
    from collections.abc import Iterable, Mapping
    from typing import Any

    from pdm.models.repositories import CandidateKey, RepositoryConfig

    from pdm_conda.environments import BaseEnvironment
    from pdm_conda.models.candidates import Candidate, FileHash
    from pdm_conda.models.requirements import Requirement


def _format_packages(packages: list[str], pretty_print=False) -> str:
    result = ""
    for i, package in enumerate(packages):
        result += f"[success]{package}[/success]" if pretty_print else package
        if i < len(packages) - 1:
            if i == len(packages) - 2:
                result += " and "
            else:
                result += ", "
    return result


class CondaRepository(BaseRepository):
    def __init__(
        self,
        sources: list[RepositoryConfig],
        environment: BaseEnvironment,
        ignore_compatibility: bool = True,
    ) -> None:
        super().__init__(sources, environment, ignore_compatibility)
        self.environment = cast(CondaEnvironment, environment)
        self._conda_resolution: dict[str, list[CondaCandidate]] = {}
        self._excluded_identifiers: set[str] = set()

    def is_conda_managed(self, requirement: Requirement, excluded_identifiers: set[str] | None = None) -> bool:
        """True if requirement is conda requirement or (not excluded and named requirement and conda as default manager
        or used by another conda requirement)

        :param requirement: requirement to evaluate
        :param excluded_identifiers: identifiers to exclude
        """
        if (
            not isinstance(self.environment, CondaEnvironment)
            or not self.environment.project.conda_config.is_initialized
        ):
            return False
        from pdm_conda.models.requirements import is_conda_managed as _is_conda_managed

        return _is_conda_managed(requirement, self.environment.project.conda_config, excluded_identifiers)

    def compatible_with_resolution(
        self,
        requirements: list[Requirement],
        resolution: dict[str, list[CondaCandidate]],
        excluded_identifiers: set[str],
    ):
        """True if all requirements are compatible with the resolution.

        :param requirements: list of requirements
        :param resolution: resolution to check
        :param excluded_identifiers: identifiers to exclude
        :return: True if all requirements are compatible with the resolution
        """
        # if any requirement not in saved candidates or incompatible candidates
        for req in requirements:
            if not self.is_conda_managed(req, excluded_identifiers):
                continue
            req = as_conda_requirement(req)
            if (key := req.conda_name) not in resolution:
                logger.info(f"Requirement {req} is not present in Conda resolution")
                return False
            for can in resolution[key]:
                if not req.is_compatible(can):
                    logger.info(f"Requirement {req} is not compatible with {can}")
                    return False
        return True

    def update_conda_resolution(
        self,
        new_requirements: list[Requirement] | None = None,
        requirements: list[Requirement] | None = None,
        resolution: dict | None = None,
        excluded_identifiers: set[str] | None = None,
    ) -> set[str]:
        """Updates the existing conda resolution if new requirements.

        :param new_requirements: new requirements to add
        :param requirements: list of requirements
        :param resolution: resolution to override existing
        :param excluded_identifiers: identifiers to exclude
        :return: list of excluded identifiers
        """
        if resolution is not None:
            self._conda_resolution = resolution
        if excluded_identifiers is not None:
            self._excluded_identifiers = excluded_identifiers
        return self._excluded_identifiers

    def get_dependencies(self, candidate: Candidate) -> tuple[list[Requirement], PySpecSet, str]:
        if isinstance(candidate, CondaCandidate):
            dependencies = list(candidate.dependencies)
            # if dep in constrains use constrain
            if candidate.constrains:
                for i, dep in enumerate(dependencies):
                    if (constrain := candidate.constrains.get(dep.identify(), None)) is not None:
                        dependencies[i] = constrain
            requires_python = PySpecSet(candidate.requires_python)
            summary = candidate.summary
        else:
            dependencies, requires_python, summary = super().get_dependencies(candidate)
        return dependencies, requires_python, summary

    def get_hashes(self, candidate: Candidate) -> list[FileHash]:
        if isinstance(candidate, CondaCandidate) and not candidate.hashes:
            logger.info(f"Fetching hashes for {candidate}")
            _candidates = conda_search(self.environment.project, candidate.req)
            if not _candidates:
                raise CondaSearchError(f"Cannot find hashes for {candidate}")

            candidate.hashes = _candidates[0].hashes
        return super().get_hashes(candidate)

    def update_hashes(self, mapping: dict[str, Candidate]):
        """Update hashes for candidates in mapping using conda create.

        :param mapping: mapping of candidates
        """
        conda_requirements = [can.req for can in mapping.values() if isinstance(can, CondaCandidate) and not can.hashes]
        if conda_requirements:
            resolution = conda_create(
                self.environment.project,
                conda_requirements,
                prefix=f"/tmp/{uuid.uuid4()}",
                dry_run=True,
            )
            for candidate in mapping.values():
                logger.info(f"Fetching hashes for {candidate}")
                if (cans := resolution.get(candidate.name, [])) and cans[0].req.is_compatible(candidate.req):
                    candidate.hashes = cans[0].hashes
                else:
                    raise CondaSearchError(f"Cannot find hashes for {candidate}")


class PyPICondaRepository(PyPIRepository, CondaRepository):
    def update_conda_resolution(
        self,
        new_requirements: list[Requirement] | None = None,
        requirements: list[Requirement] | None = None,
        resolution: dict | None = None,
        excluded_identifiers: set[str] | None = None,
    ) -> set[str]:
        excluded_identifiers = super().update_conda_resolution(
            new_requirements=new_requirements,
            requirements=requirements,
            resolution=resolution,
            excluded_identifiers=excluded_identifiers,
        )
        _requirements = [
            as_conda_requirement(req)
            for req in new_requirements or []
            if self.is_conda_managed(req, excluded_identifiers)
        ]

        if _requirements:
            try:
                _requirements += [
                    as_conda_requirement(req)
                    for req in requirements or []
                    if self.is_conda_managed(req, excluded_identifiers)
                ]
                new_resolution = conda_create(
                    self.environment.project,
                    _requirements,
                    prefix=f"/tmp/{uuid.uuid4()}",
                    dry_run=True,
                )
                conda_requirements = {r.conda_name: r for r in _requirements}
                for name, candidates in new_resolution.items():
                    req = conda_requirements.get(name, candidates[0].req)
                    key = req.conda_name
                    self._conda_resolution[key] = candidates

            except CondaResolutionError as err:
                logger.debug(err)
                if err.packages:
                    if self.environment.project.conda_config.auto_excludes:
                        logger.info(f"Adding {_format_packages(err.packages)} to excludes list")
                        return self.update_conda_resolution(
                            new_requirements,
                            requirements,
                            resolution,
                            excluded_identifiers | set(err.packages),
                        )

                    raise CandidateNotFound(
                        f"Unable to find candidates for {_format_packages(err.packages, pretty_print=True)} "
                        f"with Conda.\n"
                        f"You should add more channels or add the packages to the excludes list.",
                    ) from err
                raise
        return excluded_identifiers

    def _find_candidates(self, requirement: Requirement, minimal_version: bool) -> Iterable[Candidate]:
        if self.is_conda_managed(requirement, excluded_identifiers=self._excluded_identifiers):
            requirement = as_conda_requirement(requirement)
            candidates = self._conda_resolution.get(requirement.conda_name, [])
            candidates = [
                c.copy_with(requirement, merge_requirements=True) for c in candidates if requirement.is_compatible(c)
            ]
            candidates = list(sort_candidates(self.environment.project, candidates, minimal_version))
        else:
            if isinstance(requirement, CondaRequirement):
                requirement = requirement.as_named_requirement()
            candidates = super()._find_candidates(requirement, minimal_version)
        return candidates


class LockedCondaRepository(LockedRepository, CondaRepository):
    def _matching_keys(self, requirement: Requirement) -> Iterable[CandidateKey]:
        yield from super()._matching_keys(requirement)
        if self.is_conda_managed(requirement):
            req_id = as_conda_requirement(requirement).identify()

            for key, can in self.packages.items():
                if isinstance(can, CondaCandidate) and req_id == key[0]:
                    yield key

    def _read_lockfile(self, lockfile: Mapping[str, Any]) -> None:
        packages = lockfile.get("package", [])
        conda_packages = []
        pypi_packages = []
        for package in packages:
            if package.get("conda_managed", False):
                conda_packages.append(package)
            else:
                pypi_packages.append(package)
        super()._read_lockfile({"package": pypi_packages, **{k: v for k, v in lockfile.items() if k != "package"}})

        if conda_packages and (
            isinstance(self.environment, CondaEnvironment) and not self.environment.project.conda_config.is_initialized
        ):
            raise CondaResolutionError(
                "Conda packages found in lock file but Conda is not initialized, "
                "you should delete the lock file or initialize pdm-conda.",
            )

        for package in conda_packages:
            can = CondaCandidate.from_lock_package(package)
            can_id = self._identify_candidate(can)
            self.packages[can_id] = can
            self.candidate_info[can_id] = (
                can.dependencies_lines,
                package.get("requires_python", ""),
                package.get("summary", ""),
            )

    def _identify_candidate(self, candidate: Candidate) -> tuple:
        if isinstance(candidate, CondaCandidate):
            return candidate.identify(), candidate.version, None, False
        return super()._identify_candidate(candidate)



from __future__ import annotations

import os
from contextlib import contextmanager
from dataclasses import dataclass, field
from enum import Enum
from functools import cached_property, wraps
from pathlib import Path
from typing import TYPE_CHECKING

from pdm.exceptions import NoConfigError, ProjectError
from pdm.formats.base import make_array
from pdm.project import Config, ConfigItem
from pdm.utils import normalize_name

from pdm_conda import logger
from pdm_conda.mapping import MAPPING_DOWNLOAD_DIR_ENV_VAR, MAPPING_URL, MAPPING_URL_ENV_VAR
from pdm_conda.models.requirements import parse_requirement
from pdm_conda.utils import fix_path, get_python_dir

if TYPE_CHECKING:
    from typing import Any

    from pdm.project import Project


class CondaRunner(str, Enum):
    CONDA = "conda"
    MAMBA = "mamba"
    MICROMAMBA = "micromamba"


class CondaSolver(str, Enum):
    CONDA = "conda"
    MAMBA = "libmamba"


CONFIGS = [
    ("active", ConfigItem("Use pdm-conda plugin", True, env_var="PDM_CONDA_ACTIVE")),
    ("runner", ConfigItem("Conda runner executable", CondaRunner.CONDA.value, env_var="PDM_CONDA_RUNNER")),
    ("solver", ConfigItem("Solver to use for Conda resolution", CondaSolver.CONDA.value, env_var="PDM_CONDA_SOLVER")),
    ("channels", ConfigItem("Conda channels to use", [])),
    (
        "as-default-manager",
        ConfigItem("Use Conda to install all possible requirements", False, env_var="PDM_CONDA_AS_DEFAULT_MANAGER"),
    ),
    (
        "batched-commands",
        ConfigItem("Execute batched install and remove commands", False, env_var="PDM_CONDA_BATCHED_COMMANDS"),
    ),
    (
        "installation-method",
        ConfigItem(
            "Whether to use hard-link or copy when installing",
            "hard-link",
            env_var="PDM_CONDA_INSTALLATION_METHOD",
        ),
    ),
    ("dependencies", ConfigItem("Dependencies to install with Conda", [])),
    ("optional-dependencies", ConfigItem("Optional dependencies to install with Conda", {})),
    ("dev-dependencies", ConfigItem("Development dependencies to install with Conda", {})),
    ("excludes", ConfigItem("Excluded dependencies from Conda", [])),
    (
        "pypi-mapping.download-dir",
        ConfigItem(
            "PyPI-Conda mapping download directory",
            Path().home() / ".pdm-conda/",
            env_var=MAPPING_DOWNLOAD_DIR_ENV_VAR,
        ),
    ),
    (
        "pypi-mapping.url",
        ConfigItem(
            "PyPI-Conda mapping url",
            MAPPING_URL,
            env_var=MAPPING_URL_ENV_VAR,
        ),
    ),
    ("custom-behavior", ConfigItem("Use pdm-conda custom behavior", False, env_var="PDM_CONDA_CUSTOM_BEHAVIOR")),
    (
        "auto-excludes",
        ConfigItem(
            "If cannot find package with Conda, add it to excludes list",
            False,
            env_var="PDM_CONDA_AUTO_EXCLUDES",
        ),
    ),
]

_CONFIG_MAP = {name: name.replace("-", "_") for (name, _) in CONFIGS}
_CONFIG_MAP |= {
    "pypi-mapping.download-dir": "mapping_download_dir",
    "pypi-mapping.url": "mapping_url",
}
_CONFIG_MAP |= {v: k for k, v in _CONFIG_MAP.items()}
_CONFIG_MAP["_excludes"] = "excludes"

CONFIGS = [(f"conda.{name}", config) for name, config in CONFIGS]
PDM_CONFIG = {
    "conda.runner": "venv.backend",
}


def is_decorated(func):
    return hasattr(func, "__wrapped__")


def is_conda_config_initialized(project: Project):
    return "conda" in project.pyproject.settings


@dataclass
class PluginConfig:
    _project: Project = field(repr=False, default=None)
    _initialized: bool = field(repr=False, default=False, compare=False)
    _dry_run: bool = field(repr=False, default=True, compare=False, init=False)
    _force_set_project_config: bool = field(repr=False, default=False, compare=False, init=False)
    _excludes: list[str] = field(repr=False, compare=False, init=False, default_factory=list)
    _excluded_identifiers: set[str] | None = field(default=None, repr=False, init=False)

    channels: list[str] = field(default_factory=list)
    runner: str = CondaRunner.CONDA
    solver: str = CondaSolver.CONDA
    active: bool = True
    as_default_manager: bool = False
    custom_behavior: bool = False
    auto_excludes: bool = False
    batched_commands: bool = False
    installation_method: str = "hard-link"
    dependencies: list[str] = field(default_factory=list, repr=False)
    optional_dependencies: dict[str, list] = field(default_factory=dict)
    dev_dependencies: dict[str, list] = field(default_factory=dict)
    mapping_download_dir: Path = field(repr=False, default=Path())
    mapping_url: str = field(repr=False, default=MAPPING_URL)

    def __post_init__(self):
        if self.runner not in list(CondaRunner):
            raise ProjectError(f"Invalid Conda runner: {self.runner}")
        if self.solver not in list(CondaSolver):
            raise ProjectError(f"Invalid Conda solver: {self.solver}")
        if self.installation_method not in ["hard-link", "copy"]:
            raise ProjectError(f"Invalid Conda installation method: {self.installation_method}")
        to_suscribe = [
            (self._project.pyproject._data, "update"),
            (self._project.pyproject, "write"),
            (self._project.pyproject, "reload"),
        ]
        for obj, name in to_suscribe:
            func = getattr(obj, name)
            if not is_decorated(func):
                setattr(obj, name, self.suscribe(self, func))
        if not self.is_initialized:
            self.is_initialized = is_conda_config_initialized(self._project)

        self._dry_run = False

    def __setattr__(self, name: str, value: Any) -> None:
        super().__setattr__(name, value)
        # if plugin config is set then maybe update pyproject settings
        if (
            ((not name.startswith("_")) or name == "_excludes")
            and name != "active"
            and not isinstance(getattr(type(self), name, None), property)
            and not callable(getattr(self, name))
        ):
            name = f"conda.{_CONFIG_MAP[name]}"
            name, config_item = next(filter(lambda n: name == n[0], CONFIGS))
            if not self._dry_run:
                name_path = name.split(".")
                name = name_path.pop(-1)
                config = self._project.pyproject.settings
                should_delete = value == config_item.default and not self._force_set_project_config
                for p in name_path:
                    if should_delete and p != "conda" and p not in config:
                        break
                    config = config.setdefault(p, {})

                if should_delete:
                    # if value is default and was not set before then delete it
                    if config.get(name, value) != value:
                        config.pop(name)
                    else:
                        should_delete = False
                else:
                    _value = value
                    if isinstance(value, list):
                        _value = make_array(value, multiline=len(value) > 1)
                    elif isinstance(value, Path):
                        _value = str(value)

                    config[name] = _value
                self.is_initialized |= self._project.pyproject.exists() and "conda" in self._project.pyproject.settings
                if (project_config_name := PDM_CONFIG.get(name)) is not None:
                    project_config = self._project.project_config
                    if should_delete:
                        project_config.pop(project_config_name, None)
                    else:
                        project_config[project_config_name] = value
            if config_item.env_var:
                os.environ.setdefault(config_item.env_var, str(value))

    def reload(self):
        _conf = self.load_config(self._project)
        with self.dry_run():
            for k, v in _conf.__dict__.items():
                if not callable(v) and k not in ("_project", "_dry_run") and getattr(self, k) != v:
                    setattr(self, k, v)

    @staticmethod
    def suscribe(config, func):
        """Suscribe to function call and after executed refresh config.

        :param config: PluginConfig to refresh
        :param func: function to decorate
        """

        @wraps(func)
        def wrapper(*args, **kwargs):
            result = func(*args, **kwargs)
            config.reload()
            return result

        return wrapper

    @contextmanager
    def with_config(self, **kwargs):
        """Context manager that temporarily updates configs without updating pyproject settings."""
        configs = ["_dry_run"] + list(kwargs)
        kwargs["_dry_run"] = True and not kwargs.get("_force_set_project_config", False)
        old_values = {}
        for name in configs:
            old_values[name] = getattr(self, name)
            setattr(self, name, kwargs[name])
        try:
            yield
        finally:
            for name in reversed(configs):
                setattr(self, name, old_values[name])

    @contextmanager
    def dry_run(self):
        """Context manager that deactivates updating pyproject settings."""
        with self.with_config():
            yield

    @contextmanager
    def write_project_config(self, show_message=False):
        """Context manager that forces writing pyproject settings."""
        with self.force_set_project_config():
            yield
        self._project.pyproject.write(show_message=show_message)

    @contextmanager
    def force_set_project_config(self):
        """Context manager that forces setting pyproject settings, even default values."""
        with self.with_config(_force_set_project_config=True):
            yield

    @property
    def excluded_identifiers(self) -> set[str]:
        if self._excluded_identifiers is None:
            self._excluded_identifiers = {parse_requirement(name).identify() for name in self._excludes}
        return self._excluded_identifiers

    @cached_property
    def project_name(self) -> str | None:
        return normalize_name(self._project.name) if self._project.name else None

    @property
    def excludes(self) -> list[str]:
        return self._excludes

    @excludes.setter
    def excludes(self, value):
        excluded: set = getattr(self, "_excludes", set())
        if set(value) != excluded:
            self._excludes = sorted(value)
            self._excluded_identifiers = None

    @property
    def is_initialized(self):
        return self._initialized and self.active

    @is_initialized.setter
    def is_initialized(self, value):
        self._initialized = value

    @staticmethod
    def check_active(func):
        """Decorator that checks if the plugin is active and temporarily updates the project config."""

        @wraps(func)
        def decorator(*args, **kwargs):
            from pdm_conda.project import CondaProject

            project = None
            for arg in args:
                if isinstance(arg, CondaProject):
                    project = arg
                    break

            if project is None:
                return func(*args, **kwargs)

            config = project.conda_config
            if not config.is_initialized:
                return func(*args, **kwargs)

            project_config = project.project_config
            old_configs = {
                name: getattr(project_config, name, None)
                for name in ("python.use_venv", "python.use_pyenv", "venv.backend", "venv.in_project")
            }
            active_venv = os.environ.pop("CONDA_DEFAULT_ENV", None)
            try:
                for name, value in zip(old_configs, (True, False, config.runner, False), strict=False):
                    if getattr(old_configs, name, None) != value:
                        project_config[name] = value
                return func(*args, **kwargs)
            finally:
                for name, value in old_configs.items():
                    if value is not None:
                        if project_config[name] != value:
                            project_config[name] = value
                    elif name in project_config:
                        del project_config[name]
                if active_venv is not None:
                    os.environ["CONDA_DEFAULT_ENV"] = active_venv

        return decorator

    @contextmanager
    def with_conda_venv_location(self):
        """Context manager that ensures the PDM venv location is set to the detected Conda environment if was the
        default value.

        :return: The path to the venv location and a boolean indicating if the value was overridden
        """
        conf_name = "venv.location"
        overridden = False
        if (previous_value := self._project.config[conf_name]) == Config.get_defaults()[conf_name] and (
            conda_prefix := os.getenv("CONDA_PREFIX", None)
        ) is not None:
            venv_location = fix_path(conda_prefix)
            for parent in (venv_location, *venv_location.parents):
                if (venv_path := (parent / "envs")).is_dir():
                    logger.info(f"Detected Conda environment: {venv_path}")
                    self._project.global_config[conf_name] = str(venv_path)
                    del self._project.config
                    overridden = True
                    break
        try:
            yield fix_path(self._project.config[conf_name]), overridden
        finally:
            self._project.global_config[conf_name] = previous_value
            if overridden:
                del self._project.config

    @classmethod
    def load_config(cls, project: Project, **kwargs) -> PluginConfig:
        """Load plugin configs from project settings.

        :param project: Project
        :param kwargs: settings overwrites
        :return: plugin configs
        """

        def flatten_config(config, allowed_levels, parent_key="", result=None) -> dict:
            if result is None:
                result = {}
            for key, v in config.items():
                key = ".".join(k for k in (parent_key, key) if k)
                if isinstance(v, dict) and key in allowed_levels:
                    return flatten_config(v, allowed_levels, key, result)

                if key not in _CONFIG_MAP:
                    raise NoConfigError(key)
                result[_CONFIG_MAP[key]] = v
            return result

        config = flatten_config(project.pyproject.settings.get("conda", {}), ["pypi-mapping"])
        for n, c in CONFIGS:
            if (prop_name := _CONFIG_MAP[n[len("conda.") :]]) not in config and c.env_var:
                value = project.config[n]
                if prop_name == "mapping_download_dir":
                    value = Path(value)
                elif prop_name in (
                    "as_default_manager",
                    "batched_commands",
                    "custom_behavior",
                    "auto_excludes",
                    "active",
                ):
                    value = str(value).lower() in ("true", "1")
                config[prop_name] = value
        config |= kwargs
        excludes = config.pop("excludes", None)
        plugin_config = PluginConfig(_project=project, **config)
        if excludes is not None:
            plugin_config.excludes = excludes
        return plugin_config

    def command(self, cmd="install", use_project_env: bool = True):
        """Get runner command args.

        :param cmd: command, install by default
        :param use_project_env: use project env or not
        :return: args list
        """
        runner = self.runner
        if cmd == "remove" and runner == CondaRunner.MAMBA:
            runner = CondaRunner.CONDA
        if isinstance(runner, CondaRunner):
            runner = runner.value

        _command = [runner, *cmd.split(" ")]

        if self.runner != CondaRunner.CONDA and cmd == "search":
            _command.insert(1, "repoquery")
        if cmd in ("install", "remove", "create", "env remove"):
            _command.append("-y")
        if cmd in ("install", "create") or (cmd == "search" and self.runner == CondaRunner.MICROMAMBA):
            _command.append("--strict-channel-priority")
        if (
            self.runner in (CondaRunner.CONDA, CondaRunner.MAMBA)
            and self.solver == CondaSolver.MAMBA
            and cmd in ("create", "install")
        ):
            _command += ["--solver", CondaSolver.MAMBA.value]
        if use_project_env and cmd.split(" ")[0] not in ("search", "env", "info"):
            _command += ["--prefix", str(get_python_dir(fix_path(self._project.environment.interpreter.path)))]
        return _command



from __future__ import annotations

import dataclasses
import re
from copy import copy
from pathlib import Path
from typing import TYPE_CHECKING, cast
from urllib.parse import urlparse

from pdm.environments import BaseEnvironment
from pdm.models.candidates import Candidate, PreparedCandidate
from pdm.models.setup import Setup
from unearth import Link

from pdm_conda.models.requirements import CondaRequirement, as_conda_requirement, parse_conda_version, parse_requirement
from pdm_conda.models.setup import CondaSetupDistribution

if TYPE_CHECKING:
    from importlib.metadata import Distribution
    from typing import Any

    from pdm.models.candidates import FileHash
    from pdm.models.reporter import BaseReporter

    from pdm_conda.models.requirements import Requirement


def parse_channel(channel_url: str) -> str:
    """Parse channel from channel url.

    :param channel_url: channel url from package
    :return: channel
    """
    channel = urlparse(channel_url).path
    if channel.startswith("/"):
        channel = channel[1:]
    return channel


@dataclasses.dataclass
class CondaPreparedCandidate(PreparedCandidate):
    candidate: CondaCandidate

    def get_dependencies_from_metadata(self) -> list[str]:
        """Get the dependencies of a candidate from pre-fetched package.

        :return: list of dependencies
        """
        return [d.as_line(as_conda=True, with_build_string=True) for d in self.candidate.dependencies]

    def prepare_metadata(self, force_build: bool = False) -> Distribution:
        # if conda candidate get setup from package
        return self.candidate.distribution


class CondaCandidate(Candidate):
    def __init__(
        self,
        req: Requirement,
        name: str | None = None,
        version: str | None = None,
        link: Link | None = None,
        dependencies: list[str] | None = None,
        constrains: list[str] | None = None,
        build_string: str | None = None,
        build_number: int = 0,
        timestamp: int = 0,
        channel: str | None = None,
        track_feature: str = "",
    ):
        super().__init__(req, name, version, link)
        self._req = cast(CondaRequirement, req)  # type: ignore
        self._preferred = None
        self._prepared: CondaPreparedCandidate | None = None
        self.conda_version = version
        self.version = parse_conda_version(version)
        dependencies = dependencies or []
        if self.req.extras:
            dependencies.append(
                self.req.as_pinned_version(self.version).as_line(with_build_string=True, with_channel=True),
            )
        self.dependencies: list[CondaRequirement] = [
            cast(CondaRequirement, parse_requirement(f"conda:{r}")) for r in dependencies
        ]
        self.constrains: dict[str, CondaRequirement] = {}
        self.hashes: list[FileHash] = (
            [
                {
                    "url": self.link.url_without_fragment,
                    "file": "",
                    "hash": f"{self.link.hash_name}:{self.link.hash}",
                },
            ]
            if self.link is not None
            else []
        )
        for r in constrains or []:
            c = cast(CondaRequirement, parse_requirement(f"conda:{r}"))
            self.constrains[str(c.conda_name)] = c
        self.build_string = build_string
        self.build_number = build_number
        self.timestamp = timestamp
        self.channel = channel
        self.track_feature = track_feature

    def copy_with(self, requirement: Requirement, merge_requirements: bool = True) -> Candidate:
        can = copy(self)
        if isinstance(requirement, CondaRequirement) and merge_requirements:
            requirement.is_python_package &= can.req.is_python_package
            requirement.version_mapping |= can.req.version_mapping
        can.req = requirement
        if can._prepared:
            can._prepared.req = can.req
        return can

    @property
    def req(self):
        return self._req

    @req.setter
    def req(self, value):
        self._req = as_conda_requirement(value)

    @property
    def distribution(self):
        return CondaSetupDistribution(
            Setup(
                name=self.name,
                summary="",
                version=self.version,
                install_requires=self.dependencies_lines,
                python_requires=self.requires_python,
            ),
        )

    @property
    def dependencies_lines(self):
        return [dep.as_line(as_conda=True, with_build_string=True, with_channel=True) for dep in self.dependencies]

    def as_lockfile_entry(self, project_root: Path) -> dict[str, Any]:
        result = super().as_lockfile_entry(project_root)
        result["conda_managed"] = True
        if self.link is None:
            raise ValueError("Uninitialized conda requirement")
        result["channel"] = self.channel
        if self.build_string is not None:
            result["build_string"] = self.build_string
        result["build_number"] = self.build_number
        if self.track_feature:
            result["track_feature"] = self.track_feature
        if self.constrains:
            result["constrains"] = [c.as_line(with_build_string=True) for c in self.constrains.values()]
        result["version"] = self.conda_version
        return result

    def prepare(self, environment: BaseEnvironment, reporter: BaseReporter | None = None) -> PreparedCandidate:
        """Prepare the candidate for installation."""
        if self._prepared is None:
            self._prepared = CondaPreparedCandidate(self, environment)
        return self._prepared

    @classmethod
    def from_lock_package(cls, package: dict) -> CondaCandidate:
        """Create conda candidate from lockfile package.

        :param package: lockfile package
        :return: conda candidate
        """
        requires_python = package.get("requires_python", "")
        dependencies = package.get("dependencies", [])
        if requires_python:
            dependencies.append(f"python {requires_python}")
        corrections = {"depends": dependencies}
        for file in package.get("files", []):
            if file.get("hash"):
                hash_name, _hash = file["hash"].split(":")
                corrections[hash_name] = _hash
                corrections["url"] = file["url"]
                break
        return CondaCandidate.from_conda_package(package | corrections)

    @classmethod
    def from_conda_package(cls, package: dict, requirement: CondaRequirement | None = None) -> CondaCandidate:
        """Create conda candidate from conda package.

        :param package: conda package
        :param requirement: conda requirement associated with conda package
        :return: conda candidate
        """
        dependencies: list = package["depends"] or []
        requires_python = None
        to_delete = []
        for d in dependencies:
            if d.startswith("__"):
                to_delete.append(d)
            elif match := re.match(r"python( .+|$)", d):
                to_delete.append(d)
                if requires_python is None:
                    requires_python = match.group(1).strip().split(" ")[0] or "*"
        for d in to_delete:
            dependencies.remove(d)
        hashes = {h: package[h] for h in ["md5"] if h in package}
        url = package["url"]
        for k, v in hashes.items():
            url += f"#{k}={v}"
        name, version = package["name"], package["version"]
        build_string = package.get("build", package.get("build_string", ""))
        channel = parse_channel(package["channel"])
        marker = package.get("marker")
        if requirement is not None:
            requirement = as_conda_requirement(copy(requirement))
            requirement.version_mapping.update({parse_conda_version(version): version})
        else:
            _line = f"conda:{name}"
            if extras := package.get("extras", []):
                _line += f"[{','.join(extras)}]"
            _line += f" {version} {build_string}"
            requirement = parse_requirement(_line)

        if marker and not requirement.marker:
            requirement.marker = parse_requirement(f"{requirement.name} ; {marker}").marker
        assert requirement is not None
        requirement.is_python_package = requires_python is not None
        requirement.groups = package.get("groups", requirement.groups)
        return CondaCandidate(
            req=requirement,
            name=name,
            version=version,
            link=Link(
                url,
                requires_python=requires_python,
                comes_from=channel,
            ),
            channel=channel,
            dependencies=dependencies,
            constrains=package.get("constrains") or [],
            build_string=build_string,
            build_number=package.get("build_number", 0),
            track_feature=package.get("track_feature", ""),
            timestamp=package.get("timestamp", 0),
        )

    def __str__(self) -> str:
        if self.req.is_named:
            return f"{self.name}@{self.conda_version}"
        return super().__str__()

    def format(self) -> str:
        """Format for output."""
        return f"[req]{self.name}[/] [warning]{self.conda_version}[/]"



from __future__ import annotations

import os
import re
from pathlib import Path

from pdm.cli import utils
from pdm.installers import synchronizers
from pdm.models import candidates, repositories, requirements, working_set


def normalize_name(name: str, lowercase: bool = True) -> str:
    """Normalize name and keep `_`.

    :param name: name to normalize
    :param lowercase: if true lowercase name
    :return: normalized name
    """
    name = re.sub(r"[^A-Za-z0-9._]+", "-", name)
    return name.lower() if lowercase else name


for m in [utils, synchronizers, candidates, requirements, repositories, working_set]:
    m.normalize_name = normalize_name


def fix_path(path: str | Path) -> Path:
    """Fix path for windows."""
    path = re.sub(r"<(?:\$?env:)?([^>]+)>", r"$\1", str(path))
    path = re.sub(r"\${2,}", "$", path)
    path = os.path.expandvars(path)
    path = Path(path).expanduser()
    assert "$" not in str(path), f"Could not expand all environment variables from {path}"
    return path


def get_python_dir(path: Path) -> Path:
    """Get the python directory for given interpreter path."""
    if path.parent.name == "bin" and path.name.startswith("python"):
        return path.parents[1]
    if path.name == "python.exe":
        return path.parent
    return path



from __future__ import annotations

from typing import cast, TYPE_CHECKING

from pdm.cli.commands.venv.backends import BACKENDS, CondaBackend as BackendBase

from pdm_conda.cli.utils import ensure_logger
from pdm_conda.conda import conda_create, conda_env_remove
from pdm_conda.models.config import CondaRunner, PluginConfig
from pdm_conda.models.requirements import parse_requirement
from pdm_conda.project import CondaProject

if TYPE_CHECKING:
    from pdm.project import Project
    from pathlib import Path


class CondaBackend(BackendBase):
    def __init__(self, project: Project, python: str | None) -> None:
        super().__init__(project, python)
        self.project = cast(CondaProject, project)

    @PluginConfig.check_active
    def create(
        self,
        name: str | None = None,
        args: tuple[str, ...] = (),
        force: bool = False,
        in_project: bool = False,
        prompt: str | None = None,
        with_pip: bool = False,
    ) -> Path:
        with ensure_logger(self.project, "conda_create"):
            return super().create(name, args, force, in_project, prompt, with_pip)

    @PluginConfig.check_active
    def get_location(self, name: str | None) -> Path:
        with self.project.conda_config.with_conda_venv_location() as (venv_location, _):
            if conda_name := (name is not None and name.startswith("conda:")):
                name = name[6:]
            if conda_name:
                location = venv_location / name
            else:
                location = super().get_location(name)
            return location

    @PluginConfig.check_active
    def _ensure_clean(self, location: Path, force: bool = False) -> None:
        if self.project.conda_config.is_initialized and location.is_dir() and force:
            conda_env_remove(self.project, prefix=location)
        super()._ensure_clean(location, force)

    @PluginConfig.check_active
    def perform_create(self, location: Path, args: tuple[str, ...], prompt: str | None = None) -> None:
        if not self.project.conda_config.is_initialized:
            return super().perform_create(location, args, prompt)
        if self.python:
            python_ver = self.python
        else:
            python = self._resolved_interpreter
            python_ver = f"{python.major}.{python.minor}"

        requirements = [parse_requirement(f"conda:python={python_ver}")]
        for arg in args:
            if arg.startswith("-"):
                break
            requirements.append(parse_requirement(f"conda:{arg}"))
        conda_create(self.project, requirements=requirements, prefix=location, fetch_candidates=False)


BACKENDS = cast(dict, BACKENDS)
for runner in CondaRunner:
    BACKENDS[runner.value] = CondaBackend



from pdm.project import Project
from pdm.signals import post_lock

from pdm_conda.project import CondaProject


@post_lock.connect
def on_post_lock(project: Project, *args, dry_run: bool, **kwargs):
    """Write the pyproject.toml file after the lock file is generated if the project is a CondaProject and auto_excludes
    is configured.

    :param project: PDM project
    :param dry_run: whether it is a dry run
    """
    if isinstance(project, CondaProject):
        config = project.conda_config
        if not dry_run and config.is_initialized and config.auto_excludes and config.excluded_identifiers:
            project.pyproject.write(show_message=False)



from __future__ import annotations

from typing import TYPE_CHECKING, cast

from pdm.cli.commands.update import Command as BaseCommand
from pdm.models.specifiers import get_specifier

from pdm_conda.models.config import PluginConfig
from pdm_conda.models.requirements import CondaRequirement
from pdm_conda.project import CondaProject

if TYPE_CHECKING:
    import argparse

    from pdm_conda.project import Project


class Command(BaseCommand):
    description = BaseCommand.__doc__
    name = "update"

    @PluginConfig.check_active
    def handle(self, project: Project, options: argparse.Namespace) -> None:
        super().handle(project=project, options=options)
        project = cast(CondaProject, project)
        if (
            project.conda_config.is_initialized
            and project.conda_config.custom_behavior
            and not options.dry_run
            and options.save_strategy
        ):
            groups = set(project.iter_groups(dev=False))
            dev_groups = set(project.iter_groups(dev=True)) - groups
            candidates = project.locked_repository.all_candidates
            requirements = {}
            for i, group in enumerate(groups | dev_groups):
                group_requirements = {}
                for identifier, req in project.get_dependencies(group).items():
                    can = candidates.get(identifier, None)
                    if can is None:
                        continue
                    updated_req = can.req

                    if req.is_named:
                        version = next(s.version for s in updated_req.specifier)
                        if options.save_strategy == "minimum":
                            req.specifier = get_specifier(f">={version}")
                        elif options.save_strategy == "compatible":
                            req.specifier = get_specifier(f"~={version}")
                        elif options.save_strategy == "exact":
                            req.specifier = get_specifier(f"=={version}")
                        if isinstance(req, CondaRequirement) and isinstance(updated_req, CondaRequirement):
                            req.is_python_package = updated_req.is_python_package

                        group_requirements[identifier] = req
                    elif options.save_strategy == "exact":
                        group_requirements[identifier] = updated_req
                if group_requirements:
                    requirements[(group, i >= len(groups))] = group_requirements

            num_groups = len(requirements)
            for i, ((group, is_dev), group_requirements) in enumerate(requirements.items()):
                project.add_dependencies(
                    group_requirements,
                    to_group=group,
                    dev=is_dev,
                    write=(write := i == num_groups - 1),
                    show_message=write,
                )



from __future__ import annotations

from typing import TYPE_CHECKING

from pdm.installers import InstallManager

from pdm_conda.conda import conda_install, conda_uninstall
from pdm_conda.models.candidates import CondaCandidate
from pdm_conda.models.setup import CondaSetupDistribution

if TYPE_CHECKING:
    from importlib.metadata import Distribution

    from pdm_conda.environments import BaseEnvironment
    from pdm_conda.models.candidates import Candidate


class CondaInstallManager(InstallManager):
    def __init__(
        self,
        environment: BaseEnvironment,
        *,
        use_install_cache: bool = False,
        rename_pth: bool = False,
    ) -> None:
        super().__init__(environment, use_install_cache=use_install_cache, rename_pth=rename_pth)
        self._batch_install_queue: dict[str, str] = {}
        self._batch_install_expected: set[str] = set()
        self._batch_uninstall_queue: dict[str, str] = {}
        self._batch_uninstall_expected: set[str] = set()

    def prepare_batch_operations(self, to_install: set[str], to_uninstall: set[str]):
        self._batch_install_expected = to_install
        self._batch_uninstall_expected = to_uninstall

    def _run_with_conda(
        self,
        conda_func,
        new_requirement: str,
        queue: dict[str, str],
        expected: set[str],
        op_name: str = "",
    ):
        if should_run := (new_requirement not in expected):
            queue = {new_requirement: op_name or new_requirement}
        else:
            queue[new_requirement] = op_name or new_requirement
            should_run = set(queue) == expected

        if should_run:
            conda_func(self.environment.project, list(queue.values()), no_deps=True)

    def install(self, candidate: Candidate) -> Distribution:
        """Install candidate, use conda if conda package else default installer.

        :param candidate: candidate to install
        """
        if isinstance(candidate, CondaCandidate):
            self._run_with_conda(
                conda_install,
                candidate.name,
                self._batch_install_queue,
                self._batch_install_expected,
                f"{candidate.link.url_without_fragment}#{candidate.link.hash}",
            )
            return candidate.distribution

        return super().install(candidate)

    def uninstall(self, dist: Distribution) -> None:
        """Uninstall distribution, use conda if conda package else default uninstaller.

        :param dist: distribution to uninstall
        """
        if isinstance(dist, CondaSetupDistribution):
            self._run_with_conda(
                conda_uninstall,
                dist.name,
                self._batch_uninstall_queue,
                self._batch_uninstall_expected,
            )
        else:
            super().uninstall(dist)

    def overwrite(self, dist: Distribution, candidate: Candidate) -> None:
        """Overwrite distribution with candidate, uninstall and install with conda if conda package else default
        overwrite.

        :param dist: distribution to uninstall
        :param candidate: candidate to install
        """
        if isinstance(candidate, CondaCandidate) or isinstance(dist, CondaSetupDistribution):
            self.uninstall(dist)
            self.install(candidate)
        else:
            super().overwrite(dist, candidate)



from __future__ import annotations

from typing import cast, TYPE_CHECKING

from pdm.cli.commands.venv.list import ListCommand as BaseCommand

from pdm_conda.cli.utils import ensure_logger
from pdm_conda.models.config import PluginConfig

if TYPE_CHECKING:
    import argparse
    from pdm.project import Project


class ListCommand(BaseCommand):
    description = BaseCommand.__doc__

    @PluginConfig.check_active
    def handle(self, project: Project, options: argparse.Namespace) -> None:
        from pdm_conda.project import CondaProject

        project = cast(CondaProject, project)
        with ensure_logger(project, "venv_list"):
            super().handle(project, options)



from __future__ import annotations

import uuid
from collections import ChainMap
from pathlib import Path
from typing import TYPE_CHECKING

from pdm.models.in_process import get_sys_config_paths
from pdm.models.specifiers import PySpecSet

from pdm_conda.conda import conda_create, conda_info, conda_list
from pdm_conda.environments.python import PythonEnvironment
from pdm_conda.project import CondaProject
from pdm_conda.utils import fix_path, get_python_dir

if TYPE_CHECKING:
    from pdm.models.working_set import WorkingSet

    from pdm_conda.models.requirements import CondaRequirement, Requirement
    from pdm_conda.project import Project


class CondaEnvironment(PythonEnvironment):
    project: CondaProject

    def __init__(self, project: Project) -> None:
        super().__init__(project)
        if self.project.conda_config.is_initialized:
            self.python_requires &= PySpecSet(f"=={self.interpreter.version}")
            self.prefix = str(get_python_dir(fix_path(self.interpreter.path)))
        self._virtual_packages: set[CondaRequirement] | None = None
        self._platform: str | None = None
        self._default_channels: list[str] | None = None
        self._base_env: Path | None = None
        self._env_dependencies: dict[str, Requirement] | None = None

    @property
    def virtual_packages(self) -> set[CondaRequirement]:
        self._check_update_info(self._virtual_packages)
        return self._virtual_packages  # type: ignore

    @property
    def platform(self) -> str:
        self._check_update_info(self._platform)
        return self._platform  # type: ignore

    @property
    def default_channels(self) -> list[str]:
        self._check_update_info(self._default_channels)
        return self._default_channels  # type: ignore

    def _check_update_info(self, prop):
        if prop is None:
            self._get_conda_info()

    def _get_conda_info(self):
        info = conda_info(self.project)
        self._virtual_packages = info["virtual_packages"]
        self._platform = info["platform"]
        self._default_channels = info["channels"]

    def get_paths(self, dist_name: str | None = None) -> dict[str, str]:
        if self.project.conda_config.is_initialized:
            paths = get_sys_config_paths(
                str(fix_path(self.interpreter.executable)),
                {k: self.prefix for k in ("base", "platbase", "installed_base")},
                kind="prefix",
            )
            paths.setdefault("prefix", self.prefix)
            paths["headers"] = paths["include"]
            return paths
        return super().get_paths(dist_name)

    def get_working_set(self) -> WorkingSet:
        """Get the working set based on local packages directory, include Conda managed packages."""
        working_set = super().get_working_set()
        if self.project.conda_config.is_initialized:
            dist_map = working_set._dist_map | conda_list(self.project)
            working_set._dist_map = dist_map
            shared_map = getattr(working_set, "_shared_map", {})
            working_set._iter_map = ChainMap(dist_map, shared_map)
        return working_set

    @property
    def env_dependencies(self) -> dict[str, Requirement]:
        if self._env_dependencies is None:
            self._env_dependencies = {}

            working_set = conda_list(self.project)
            dependencies = ["python"]
            if (runner := self.project.conda_config.runner) in working_set:
                dependencies.append(runner)
            self._env_dependencies = conda_create(
                self.project,
                [working_set[d].req for d in dependencies],
                prefix=f"/tmp/{uuid.uuid4()}",
                dry_run=True,
            )

        return self._env_dependencies  # type: ignore



from __future__ import annotations

from functools import cached_property
from typing import TYPE_CHECKING, cast

from pdm.exceptions import PdmUsageError, ProjectError
from pdm.models.python import PythonInfo
from pdm.project import Project
from pdm.project.lockfile import Lockfile
from pdm.utils import get_venv_like_prefix
from tomlkit.items import Array

from pdm_conda import logger
from pdm_conda.models.config import PluginConfig
from pdm_conda.models.requirements import CondaRequirement, as_conda_requirement, is_conda_managed, parse_requirement
from pdm_conda.project.project_file import PyProject

if TYPE_CHECKING:
    from collections.abc import Iterable
    from pathlib import Path

    from pdm.core import Core
    from pdm.environments import BaseEnvironment
    from pdm.models.repositories import LockedRepository
    from pdm.models.requirements import Requirement
    from pdm.resolver.providers import BaseProvider


class CondaProject(Project):
    def __init__(
        self,
        core: Core,
        root_path: str | Path | None,
        is_global: bool = False,
        global_config: str | Path | None = None,
    ) -> None:
        from pdm_conda.installers.manager import CondaInstallManager
        from pdm_conda.installers.synchronizers import CondaSynchronizer
        from pdm_conda.models.repositories import LockedCondaRepository, PyPICondaRepository
        from pdm_conda.resolvers import CondaResolver

        super().__init__(core, root_path, is_global, global_config)
        self.core.repository_class = PyPICondaRepository
        self.core.install_manager_class = CondaInstallManager
        self.core.synchronizer_class = CondaSynchronizer
        self.core.resolver_class = CondaResolver
        self.locked_repository_class = LockedCondaRepository
        self._conda_mapping: dict[str, str] = {}
        self._pypi_mapping: dict[str, str] = {}
        self.conda_config = PluginConfig.load_config(self)
        self._is_distribution: bool | None = None
        self._base_env: Path | None = None

    @property
    def virtual_packages(self) -> set[CondaRequirement]:
        from pdm_conda.environments import CondaEnvironment

        if isinstance(self.environment, CondaEnvironment):
            return self.environment.virtual_packages
        return set()

    @property
    def platform(self) -> str:
        from pdm_conda.environments import CondaEnvironment

        if isinstance(self.environment, CondaEnvironment):
            return self.environment.platform
        return ""

    @property
    def default_channels(self) -> list[str]:
        from pdm_conda.environments import CondaEnvironment

        if isinstance(self.environment, CondaEnvironment):
            return self.environment.default_channels
        return []

    @property
    def base_env(self) -> Path:
        if self._base_env is None:
            from pdm_conda.conda import conda_base_path

            self._base_env = conda_base_path(self)
        return self._base_env

    @property
    def locked_repository(self) -> LockedRepository:
        try:
            lockfile = self.lockfile._data.unwrap()
        except ProjectError:
            lockfile = {}

        return self.locked_repository_class(lockfile=lockfile, sources=self.sources, environment=self.environment)  # type: ignore

    @Project.python.setter
    @PluginConfig.check_active
    def python(self, value: PythonInfo) -> None:
        if not self.conda_config.is_initialized:
            Project.python.fset(self, value)
            return

        self._python = value
        self._saved_python = value.path.as_posix()
        self.environment = None

    @cached_property
    def pyproject(self) -> PyProject:
        return PyProject(self.root / self.PYPROJECT_FILENAME, ui=self.core.ui)

    @property
    def lockfile(self) -> Lockfile:
        if self._lockfile is None:
            self.set_lockfile(self.root / self.LOCKFILE_FILENAME)
        return self._lockfile

    def set_lockfile(self, path: str | Path) -> None:
        self._lockfile = Lockfile(path, ui=self.core.ui)
        # conda don't produce cross-platform locks
        if self.conda_config.is_initialized and not self._lockfile.empty():
            self._lockfile._data.setdefault("metadata", {})["cross_platform"] = False

    def _get_conda_info(self):
        from pdm_conda.conda import conda_info

        info = conda_info(self)
        self._virtual_packages = info["virtual_packages"]
        self._platform = info["platform"]
        self._default_channels = info["channels"]

    def get_conda_pyproject_dependencies(self, group: str, dev: bool = False, set_defaults=False) -> list[str]:
        """Get the conda dependencies array in the pyproject.toml."""

        def _getter(conf, name, default, set_defaults=False):
            return (conf.setdefault if set_defaults else conf.get)(name, default)

        settings = _getter(self.pyproject.settings, "conda", {}, set_defaults)
        if group == "default":
            group = "dependencies"
        else:
            name = "optional" if not dev else "dev"
            settings = _getter(settings, f"{name}-dependencies", {}, set_defaults)
        return _getter(settings, group, [], set_defaults)

    def iter_groups(self, dev: bool = True) -> Iterable[str]:
        groups = set(super().iter_groups())
        config = self.conda_config
        if not config.is_initialized:
            return groups
        for is_dev, deps in ((False, config.optional_dependencies), (True, config.dev_dependencies)):
            if deps and (dev or not is_dev):
                groups.update(deps.keys())
        if not dev:
            for group in self.pyproject.settings.get("dev-dependencies", {}):
                groups.remove(group)
        return groups

    def get_dependencies(self, group: str | None = None) -> dict[str, Requirement]:
        config = self.conda_config
        if not config.is_initialized:
            return super().get_dependencies(group)

        group = group or "default"
        dev = group not in config.optional_dependencies
        try:
            result = super().get_dependencies(group)
        except PdmUsageError:
            result = {}

        if group in config.optional_dependencies and group in config.dev_dependencies:
            self.core.ui.echo(
                f"The {group} group exists in both [optional-dependencies] "
                "and [dev-dependencies], the former is taken.",
                err=True,
                style="warning",
            )
        deps = self.get_conda_pyproject_dependencies(group, dev)

        for line in deps:
            req = parse_requirement(f"conda:{line}")
            req.groups = [group]
            # search for package with extras to remove it
            pypi_req = next((v for v in result.values() if v.conda_name == req.conda_name), None)
            if pypi_req is not None:
                result.pop(pypi_req.identify())
                if not req.specifier:
                    req.specifier = pypi_req.specifier
                if pypi_req.marker:
                    req.marker = pypi_req.marker
                if pypi_req.extras:
                    req.extras = pypi_req.extras
                req.groups = pypi_req.groups
            result[req.identify()] = req

        if self.conda_config.as_default_manager:
            for k in list(result):
                if is_conda_managed(req := result[k], config):
                    result[k] = as_conda_requirement(req)

        return result

    def add_dependencies(
        self,
        requirements: dict[str, Requirement],
        to_group: str = "default",
        dev: bool = False,
        show_message: bool = True,
        write: bool = True,
    ) -> None:
        conda_requirements = {n: r for n, r in requirements.items() if isinstance(r, CondaRequirement)}
        requirements = {n: r for n, r in requirements.items() if n not in conda_requirements} | {
            n: r.as_named_requirement() for n, r in conda_requirements.items() if r.is_python_package
        }
        if self.conda_config.is_initialized:
            if self.conda_config.as_default_manager:
                conda_requirements = {
                    n: r
                    for n, r in conda_requirements.items()
                    if not r.is_python_package or r.channel or r.build_string
                }
            if conda_requirements:
                deps = self.get_conda_pyproject_dependencies(to_group, dev, set_defaults=True)
                python_deps, _ = self.use_pyproject_dependencies(to_group, dev)
                cast(Array, deps).multiline(True)
                for name, dep in conda_requirements.items():
                    matched_index = next((i for i, r in enumerate(deps) if dep.matches(f"conda:{r}")), None)
                    req = dep.as_line(with_channel=True)
                    if matched_index is None:
                        deps.append(req)
                    else:
                        deps[matched_index] = req
                    if name not in requirements:
                        matched_index = next((i for i, r in enumerate(python_deps) if dep.matches(r)), None)
                        if matched_index is not None:
                            python_deps.pop(matched_index)
        else:
            assert not conda_requirements, "Conda is not initialized but conda requirements are present."

        super().add_dependencies(requirements, to_group, dev, show_message, write=write)

    @PluginConfig.check_active
    def get_environment(self) -> BaseEnvironment:
        if not self.conda_config.is_initialized:
            return super().get_environment()

        if not get_venv_like_prefix(self.python.executable)[1]:
            logger.debug("Conda environment not detected.")
            return super().get_environment()
        if not self.config["python.use_venv"]:
            raise ProjectError("python.use_venv is required to use Conda.")
        from pdm_conda.environments import CondaEnvironment

        return CondaEnvironment(self)

    def get_provider(
        self,
        strategy: str = "all",
        tracked_names: Iterable[str] | None = None,
        for_install: bool = False,
        ignore_compatibility: bool = True,
        direct_minimal_versions: bool = False,
    ) -> BaseProvider:
        if not self.conda_config.is_initialized:
            return super().get_provider(strategy, tracked_names, for_install, ignore_compatibility)

        from pdm_conda.resolver.providers import BaseProvider, CondaBaseProvider

        kwargs = {"direct_minimal_versions": direct_minimal_versions}
        provider = super().get_provider(strategy, tracked_names, for_install, ignore_compatibility, **kwargs)
        if isinstance(provider, BaseProvider) and not isinstance(provider, CondaBaseProvider):
            kwargs["locked_candidates"] = provider.locked_candidates
            return CondaBaseProvider(provider.repository, **kwargs)  # type: ignore[arg-type]
        return provider

    @property
    def is_distribution(self) -> bool:
        if self._is_distribution is not None:
            return self._is_distribution
        return super().is_distribution

    @is_distribution.setter
    def is_distribution(self, value: bool | None):
        self._is_distribution = value

    @PluginConfig.check_active
    def find_interpreters(
        self,
        python_spec: str | None = None,
        search_venv: bool | None = None,
    ) -> Iterable[PythonInfo]:
        if not self.conda_config.is_initialized:
            return super().find_interpreters(python_spec, search_venv)
        else:
            roots = set()
            if self.base_env:
                roots.add(self.base_env)
            for i in super().find_interpreters(python_spec, search_venv):
                if (venv := i.get_venv()) is not None and venv.is_conda:
                    if (root := venv.root) not in roots:
                        roots.add(root)
                        yield i
                else:
                    yield i



from __future__ import annotations

import argparse

from pdm.cli.commands.venv import Command as BaseCommand

from pdm_conda.cli.commands.venv.create import CreateCommand
from pdm_conda.cli.commands.venv.list import ListCommand


class Command(BaseCommand):
    description = BaseCommand.__doc__
    name = "venv"

    def add_arguments(self, parser: argparse.ArgumentParser) -> None:
        super().add_arguments(parser)
        subparser = parser._actions[-1]
        CreateCommand.register_to(subparser, "create")
        ListCommand.register_to(subparser, "list")



from __future__ import annotations

import contextlib
import json
import os
import re
import subprocess
from functools import cache
from pathlib import Path
from shutil import which
from tempfile import TemporaryDirectory, gettempdir
from typing import TYPE_CHECKING

from pdm.cli.commands.venv.backends import VirtualenvCreateError
from pdm.exceptions import InstallationError, PdmException, RequirementError, UninstallError
from pdm.models.finder import ReverseVersion
from pdm.models.setup import Setup
from pdm.termui import Verbosity

from pdm_conda import logger
from pdm_conda.models.candidates import CondaCandidate, parse_channel
from pdm_conda.models.conda import ChannelSorter
from pdm_conda.models.config import CondaRunner, PluginConfig
from pdm_conda.models.requirements import CondaRequirement, parse_conda_version, parse_requirement
from pdm_conda.models.setup import CondaSetupDistribution
from pdm_conda.utils import fix_path, normalize_name

if TYPE_CHECKING:
    from collections.abc import Iterable

    from pdm_conda.project import CondaProject

_conda_response_packages_res = [
    re.compile(r"(nothing provides( requested)?|^.(\s+.)?─)\s+(?P<package>\S+)"),
    re.compile(r"(nothing provides .* needed by)\s+(?P<package>.+)-\d+\.\w+\.\w+-.+$"),
]


class CondaExecutionError(PdmException):
    def __init__(self, *args, data: dict | None = None):
        super().__init__(*args)
        self.data = data or {}
        self.message = self.data.get("message", "")


class CondaResolutionError(CondaExecutionError):
    def __init__(self, *args, data: dict | None = None):
        super().__init__(*args, data=data)
        self.packages: list = self.data.get("packages", [])


class CondaSearchError(CondaExecutionError):
    pass


class CondaRunnerNotFoundError(CondaExecutionError):
    pass


@contextlib.contextmanager
def _optional_temporary_file(environment: dict | list):
    """If environment contains data then creates temporary file else yield None.

    :param environment: environment data
    :return: Temporary file or None
    """
    if environment:
        with (
            TemporaryDirectory() as temp_dir,
            Path(temp_dir).joinpath("lock" + ".yml" if isinstance(environment, dict) else ".lock").open("w+") as f,
        ):
            yield f
    else:
        yield


def run_conda(
    cmd,
    exception_cls: type[PdmException] = CondaExecutionError,
    exception_msg: str = "Error locking dependencies",
    env: dict | None = None,
    **environment,
) -> dict:
    """Optionally creates temporary environment file and run conda command.

    :param cmd: conda command
    :param exception_cls: exception to raise on error
    :param exception_msg: base message to show on error
    :param env: environment variables to use for conda
    :param environment: environment or lockfile data
    :return: conda command response
    """
    executable = which(cmd[0])
    if executable is None:
        raise CondaRunnerNotFoundError(f"Conda runner {cmd[0]} not found.")

    lockfile = environment.get("lockfile", [])
    with _optional_temporary_file(lockfile or environment) as f:
        if lockfile or environment:
            if lockfile:
                f.write("\n".join(lockfile))
            elif environment:
                for name, options in environment.items():
                    if options:
                        f.write(f"{name}:")
                        if isinstance(options, str):
                            f.write(f" {options}\n")
                            continue

                        f.write("\n")
                        for v in options:
                            f.write(f"  - {v}\n")
            f.seek(0)
            cmd += ["--file", f.name]
        logger.debug(f"cmd: {' '.join(cmd)}")
        if environment:
            logger.debug(f"env: {environment}")
        process = subprocess.run(cmd, capture_output=True, encoding="utf-8", env=env)
    if "--json" in cmd:
        try:
            out = process.stdout.strip()
            if not out.startswith("{") and not out.startswith("["):
                out = "{" + out.split("{")[-1]

            response = json.loads(out)
        except json.JSONDecodeError:
            response = {}
    else:
        response = {}
        msg = f"{process.stdout}\n" if process.stdout else ""
        if process.stderr:
            msg += process.stderr
        if msg:
            response["message"] = msg
    try:
        process.check_returncode()
    except subprocess.CalledProcessError as e:
        msg = ""
        if exception_msg:
            msg = f"{exception_msg}\n"
        kwargs = {}
        if isinstance(response, dict) and not response.get("success", False):
            if err := response.get(
                "solver_problems",
                response.get("error", response.get("message", f"{process.stderr}\n{process.stdout}")),
            ):
                if isinstance(err, str):
                    err = [err]
                msg += "\n".join(err)
            if exception_cls == CondaResolutionError:
                if "message" not in response:
                    response["message"] = process.stdout or process.stderr
                kwargs["data"] = response
        else:
            msg += process.stderr
        if environment:
            msg += f"\n{environment}"
        raise exception_cls(msg, **kwargs) from e
    return response


@cache
def _get_channel_sorter(platform: str, channels: tuple[str]) -> ChannelSorter:
    """Get channel sorter.

    :param channels: list of conda channels used to determine priority
    :param platform: env platform
    :return: channel sorter
    """
    return ChannelSorter(platform, channels)


def sort_candidates(
    project: CondaProject,
    packages: list[CondaCandidate],
    minimal_version: bool,
) -> Iterable[CondaCandidate]:
    """
    Sort candidates following mamba specification
    (https://mamba.readthedocs.io/en/latest/advanced_usage/package_resolution.html).

    :param project: PDM project
    :param packages: list of conda candidates
    :return: sorted conda candidates
    """
    if len(packages) <= 1:
        return packages
    channels_sorter = _get_channel_sorter(project.platform, tuple(project.conda_config.channels))

    def get_preference(candidate: CondaCandidate):
        return (
            not candidate.track_feature,
            ReverseVersion(candidate.version) if minimal_version else candidate.version,
            candidate.build_number,
            -channels_sorter.get_priority(candidate.channel or ""),
            candidate.timestamp,
        )

    return sorted(packages, key=get_preference, reverse=True)


def _parse_candidates(project: CondaProject, packages: list[dict], requirement=None) -> list[CondaCandidate]:
    """Convert conda packages to candidates.

    :param project: PDM project
    :param packages: conda packages
    :param requirement: requirement linked to packages
    :return: list of candidates
    """
    candidates = []
    for p in packages:
        dependencies = p.get("depends", None) or []
        valid_candidate = True
        for d in dependencies:
            if d.startswith("__"):
                d = parse_requirement(f"conda:{d}")
                if not any(d.is_compatible(v) for v in project.virtual_packages):
                    valid_candidate = False
                    break
        if valid_candidate:
            candidates.append(CondaCandidate.from_conda_package(p, requirement))

    return candidates


def _ensure_channels(
    project: CondaProject,
    channels: list[str],
    log_message: str = "No channels specified, using defaults if exist.",
) -> list[str]:
    """Ensure channels and if empty use defaults.

    :param project: PDM project
    :param channels: channels to validate
    :param log_message: log message to display if using defaults
    :return: list of channels
    """
    channels = channels or project.conda_config.channels
    if not channels:
        project.core.ui.echo(log_message, verbosity=Verbosity.DEBUG)
        channels.append("defaults")

    return list(dict.fromkeys(channels))


@cache
def _conda_search(project: CondaProject, requirement: str, channels: tuple[str], use_cache: bool = False) -> list[dict]:
    """Search conda candidates for a requirement.

    :param project: PDM project
    :param requirement: requirement
    :param channels: requirement channels
    :param use_cache: whether to use cache flag
    :return: list of conda candidates
    """
    config = project.conda_config
    command = config.command("search")
    command.append(requirement)

    for c in channels:
        command.extend(["-c", c])
    if channels:
        command.append("--override-channels")
    command.append("--json")
    if use_cache:
        command.append("-C")
        command.append("--offline")
    try:
        result = run_conda(command)
    except RequirementError as e:
        if "PackagesNotFoundError:" in str(e):
            result = {}
        else:
            raise

    if config.runner == CondaRunner.CONDA:
        packages = result.get(parse_requirement(f"conda:{requirement}").name, [])
    else:
        packages = result.get("result", {}).get("pkgs", [])
    return packages


@PluginConfig.check_active
def conda_search(
    project: CondaProject,
    requirement: CondaRequirement | str,
    channel: str | None = None,
    use_cache: bool = False,
) -> list[CondaCandidate]:
    """Search conda candidates for a requirement.

    :param project: PDM project
    :param requirement: requirement
    :param channel: requirement channel
    :param use_cache: whether to use cache flag
    :return: list of conda candidates
    """
    _requirement = requirement
    if isinstance(_requirement, CondaRequirement):
        channel = channel or _requirement.channel
        _requirement = _requirement.as_line(with_build_string=True, conda_compatible=True).replace(" ", "=")
    if "::" in _requirement:
        channel, _requirement = _requirement.split("::", maxsplit=1)
    if isinstance(requirement, str):
        requirement = parse_requirement(f"conda:{requirement}")
    channels = _ensure_channels(
        project,
        [channel] if channel else [],
        f"No channel specified for searching [req]{requirement}[/] using defaults if exist.",
    )
    packages = _conda_search(project, _requirement, tuple(channels), use_cache=use_cache)
    return _parse_candidates(project, packages, requirement)


@PluginConfig.check_active
def conda_create(
    project: CondaProject,
    requirements: Iterable[CondaRequirement],
    channels: list[str] | None = None,
    prefix: Path | str | None = None,
    name: str = "",
    dry_run: bool = False,
    fetch_candidates: bool = True,
) -> dict[str, list[CondaCandidate]]:
    """Creates environment using conda.

    :param project: PDM project
    :param requirements: conda requirements
    :param channels: requirement channels
    :param prefix: environment prefix
    :param name: environment name
    :param dry_run: don't install if dry run
    :param fetch_candidates: if True ensure ensure candidates were fetched
    """
    config = project.conda_config
    if not config.is_initialized:
        raise VirtualenvCreateError("Error creating environment, no pdm-conda configs were found on pyproject.toml.")
    candidates = {}
    channels = channels or []
    for req in requirements:
        if req.channel:
            channels.append(req.channel)
    channels = _ensure_channels(
        project,
        channels,
        "No channels specified for creating environment, using defaults if exist.",
    )
    command = config.command("create", use_project_env=False)
    command.append("--json")
    if prefix is not None:
        command.extend(["--prefix", str(fix_path(prefix))])
    elif name:
        command.extend(["--name", name])
    else:
        raise VirtualenvCreateError("Error creating environment, name or prefix must be specified.")

    env = {**os.environ}
    if dry_run:
        # momentarily change download dir to always get fetch info
        env["CONDA_PKGS_DIRS"] = str(Path(gettempdir()) / "conda_cache")
        command.append("--dry-run")

    command += list(
        {
            req.as_line(with_build_string=True, conda_compatible=True, with_channel=True).replace(" ", "="): None
            for req in requirements
        },
    )

    if channels:
        for c in channels:
            command.extend(["-c", c])
        command.append("--override-channels")

    try:
        result = run_conda(
            command,
            exception_cls=CondaResolutionError if dry_run else VirtualenvCreateError,
            exception_msg=(
                f"Error resolving requirements with {config.runner}" if dry_run else "Error creating environment"
            ),
            env=env,
        )
        if fetch_candidates:
            actions = result.get("actions", {})
            fetch_packages = {pkg["name"]: pkg for pkg in actions.get("FETCH", [])}
            packages = actions.get("LINK", [])
            for i, pkg in enumerate(packages):
                pkg = fetch_packages.get(pkg["name"], pkg)
                if any(True for n in ("constrains", "depends") if n not in pkg):
                    pkg = conda_search(
                        project,
                        f'{pkg["name"]}={pkg["version"]}={pkg["build_string"]}',
                        parse_channel(pkg["channel"]),
                    )
                packages[i] = pkg

            _requirements = {req.conda_name: req for req in requirements}
            for pkg in packages:
                # if is list of candidates then it comes from search
                if isinstance(pkg, list):
                    if pkg:
                        candidates[pkg[0].name] = pkg
                else:
                    name = pkg["name"]
                    candidates[name] = _parse_candidates(
                        project,
                        packages=[pkg],
                        requirement=_requirements.get(name),
                    )
        return candidates
    except CondaResolutionError as err:
        if not err.packages:
            failed_packages = set()
            for line in err.message.split("\n"):
                for pat in _conda_response_packages_res:
                    if (match := pat.search(line)) is not None:
                        failed_packages.add(match.group("package"))
            err.packages = list(failed_packages)
        raise


@PluginConfig.check_active
def conda_env_remove(project: CondaProject, prefix: Path | str | None = None, name: str = "", dry_run: bool = False):
    """Removes environment using conda.

    :param project: PDM project
    :param prefix: environment prefix
    :param name: environment name
    :param dry_run: don't install if dry run
    """
    config = project.conda_config
    if not config.is_initialized:
        raise VirtualenvCreateError("Error removing environment, no pdm-conda configs were found on pyproject.toml.")
    command = config.command("env remove", use_project_env=False)
    command.append("--json")
    if prefix is not None:
        command += ["--prefix", str(fix_path(prefix))]
    elif name:
        command += ["--name", name]
    else:
        raise VirtualenvCreateError("Error removing environment, name or prefix must be specified.")

    if dry_run:
        command.append("--dry-run")

    run_conda(command, exception_cls=VirtualenvCreateError, exception_msg="Error removing environment")


@PluginConfig.check_active
def conda_env_list(project: CondaProject) -> list[Path]:
    """List Conda environments.

    :param project: PDM project
    :return: list of conda environments
    """
    config = project.conda_config
    command = config.command("env list", use_project_env=False)
    command.append("--json")
    environments = run_conda(command, exception_cls=CondaExecutionError, exception_msg="Error listing environments")
    return [fix_path(env) for env in environments.get("envs", [])]


def _conda_install(
    command: list[str],
    packages: str | list[str] | None = None,
    exception_cls: type[PdmException] = InstallationError,
    dry_run: bool = False,
    explicit: bool = False,
):
    if isinstance(packages, str):
        packages = [packages]
    if not packages:
        raise exception_cls(f"No packages used for {' '.join(command[:3])} command.")
    if explicit:
        packages.insert(0, "@EXPLICIT")
    else:
        command.extend(packages)
    if dry_run:
        command.append("--dry-run")
    kwargs: dict = {}
    if explicit:
        kwargs["lockfile"] = packages
    run_conda(command + ["--json"], exception_cls=exception_cls, exception_msg="", **kwargs)


@PluginConfig.check_active
def conda_install(
    project: CondaProject,
    packages: str | list[str],
    dry_run: bool = False,
    no_deps: bool = False,
):
    """Install resolved packages using conda.

    :param project: PDM project
    :param packages: resolved packages
    :param dry_run: don't install if dry run
    :param no_deps: don't install dependencies if true
    """
    config = project.conda_config
    command = config.command("install")
    if no_deps:
        command.append("--no-deps")
        if config.runner != CondaRunner.MICROMAMBA:
            command.append("--no-update-deps")

    if config.installation_method == "copy":
        _copy = "copy"
        if config.runner == CondaRunner.MICROMAMBA:
            _copy = f"always-{_copy}"
        command.append(f"--{_copy}")

    _conda_install(command, packages, dry_run=dry_run, explicit=True)


@PluginConfig.check_active
def conda_uninstall(
    project: CondaProject,
    packages: str | list[str],
    dry_run: bool = False,
    no_deps: bool = False,
):
    """Uninstall resolved packages using conda.

    :param project: PDM project
    :param packages: resolved packages
    :param dry_run: don't uninstall if dry run
    :param no_deps: don't uninstall dependencies if true
    """
    config = project.conda_config

    command = config.command("remove")
    if no_deps:
        if config.runner == CondaRunner.MICROMAMBA:
            command.append("--no-prune")
        command.append("--force")

    _conda_install(command, packages, dry_run=dry_run, exception_cls=UninstallError)


def not_initialized_warning(project):
    project.core.ui.echo(
        "[warning]Tried to execute a conda command but no pdm-conda configs were found on pyproject.toml.[/]",
    )


@PluginConfig.check_active
def conda_info(project: CondaProject) -> dict:
    """Get conda info containing virtual packages, default channels and packages.

    :param project: PDM project
    :return: dict with conda info
    """
    config = project.conda_config
    res: dict = {"virtual_packages": set(), "platform": "", "channels": []}
    if config.is_initialized:
        cmd = config.command("info") + ["--json"]
        info = run_conda(cmd)
        if config.runner != CondaRunner.MICROMAMBA:
            virtual_packages = {"=".join(p) for p in info["virtual_pkgs"]}
        else:
            virtual_packages = set(info["virtual packages"])

        res["virtual_packages"] = {parse_requirement(f"conda:{p.replace('=', '==', 1)}") for p in virtual_packages}
        res["platform"] = info["platform"]
        res["channels"] = [parse_channel(channel) for channel in (info["channels"] or [])]
    else:
        not_initialized_warning(project)
    return res


@PluginConfig.check_active
def conda_base_path(project: CondaProject) -> Path:
    """Get conda base environment path :param project: PDM project :return: Conda base environment path."""
    config = project.conda_config
    res = Path()
    if config.is_initialized:
        cmd = config.command("info")
        if config.runner == CondaRunner.MICROMAMBA:
            cmd += ["-n", "base"]
        else:
            cmd.append("--base")
        cmd += ["--json"]
        info = run_conda(cmd)
        res = fix_path(info.get("base environment", info.get("root_prefix", res)))
    else:
        not_initialized_warning(project)
    return res


@PluginConfig.check_active
def conda_list(project: CondaProject) -> dict[str, CondaSetupDistribution]:
    """List conda installed packages.

    :param project: PDM project
    :return: packages distribution
    """
    config = project.conda_config
    distributions = {}
    if config.is_initialized:
        packages = run_conda(config.command("list") + ["--json"], exception_msg="Error listing installed packages")
        for package in packages:
            if config.runner != CondaRunner.MICROMAMBA and package.get("platform", "") == "pypi":
                continue
            name, version = package["name"], package["version"]
            distributions[normalize_name(name)] = CondaSetupDistribution(
                Setup(
                    name=name,
                    summary="",
                    version=parse_conda_version(version),
                ),
                package=package,
            )
    else:
        not_initialized_warning(project)
    return distributions



from __future__ import annotations

from typing import TYPE_CHECKING, cast

from pdm.cli.commands.lock import Command as BaseCommand
from pdm.project.lockfile import FLAG_CROSS_PLATFORM

from pdm_conda.models.config import PluginConfig
from pdm_conda.project import CondaProject

if TYPE_CHECKING:
    import argparse

    from pdm_conda.project import Project


class Command(BaseCommand):
    description = BaseCommand.__doc__
    name = "lock"

    @PluginConfig.check_active
    def handle(self, project: Project, options: argparse.Namespace) -> None:
        project = cast(CondaProject, project)
        if project.conda_config.is_initialized:
            # conda doesn't produce cross-platform locks
            options.strategy_change = [
                s for s in (options.strategy_change or []) if not s.replace("-", "_").endswith(FLAG_CROSS_PLATFORM)
            ] + [
                f"no_{FLAG_CROSS_PLATFORM}",
            ]
            if options.groups and ":all" in options.groups:
                options.groups += list(project.iter_groups(dev=True if options.dev is None else options.dev))
        super().handle(project=project, options=options)



import functools
from collections import ChainMap
from copy import copy

from pdm.environments import PythonEnvironment

from pdm_conda.mapping import pypi_to_conda
from pdm_conda.project import CondaProject
from pdm_conda.utils import normalize_name


def wrap_get_working_set(func):
    @functools.wraps(func)
    def wrapper(self, *args, **kwargs):
        res = func(self, *args, **kwargs)
        dist_map = {normalize_name(pypi_to_conda(dist.metadata["Name"])): dist for dist in res._dist_map.values()}
        res._dist_map = dist_map
        res._iter_map = ChainMap(dist_map, getattr(res, "_shared_map", {}))
        return res

    return wrapper


def wrap_init(func):
    @functools.wraps(func)
    def wrapper(self, *args, **kwargs):
        res = func(self, *args, **kwargs)
        if isinstance(self.project, CondaProject):
            self.project = copy(self.project)
            self.project.environment = self
        return res

    return wrapper


PythonEnvironment.__init__ = wrap_init(PythonEnvironment.__init__)
PythonEnvironment.get_working_set = wrap_get_working_set(PythonEnvironment.get_working_set)



from __future__ import annotations

from typing import TYPE_CHECKING

from pdm.cli.commands.list import Command as BaseCommand

from pdm_conda.cli.utils import ensure_logger
from pdm_conda.models.config import PluginConfig

if TYPE_CHECKING:
    import argparse

    from pdm_conda.project import Project


class Command(BaseCommand):
    description = BaseCommand.__doc__
    name = "list"

    @PluginConfig.check_active
    def handle(self, project: Project, options: argparse.Namespace) -> None:
        with ensure_logger(project, "list"):
            super().handle(project, options)



from __future__ import annotations

from typing import TYPE_CHECKING, cast

from pdm.cli.commands.install import Command as BaseCommand

from pdm_conda.models.config import PluginConfig
from pdm_conda.project import CondaProject

if TYPE_CHECKING:
    import argparse

    from pdm_conda.project import Project


class Command(BaseCommand):
    description = BaseCommand.__doc__
    name = "install"

    @PluginConfig.check_active
    def handle(self, project: Project, options: argparse.Namespace) -> None:
        project = cast(CondaProject, project)
        if options.groups and ":all" in options.groups:
            options.groups += list(project.iter_groups())
        super().handle(project, options)
